import bisect
import collections
from copy import deepcopy
import datetime as dt
import typing

from creme import base


__all__ = ['simulate_qa']


class Memento(collections.namedtuple('Memento', 'i x y t_expire')):

    def __lt__(self, other):
        return self.t_expire < other.t_expire


def simulate_qa(X_y: base.typing.Stream, moment: typing.Union[str, typing.Callable],
                delay: typing.Union[str, int, dt.timedelta, typing.Callable], copy: bool = True):
    """Simulate a time-ordered question and answer session.

    This method allows looping through a dataset in the order in which it arrived. Indeed, it
    usually is the case that labels arrive after features. Being able to go through a dataset in
    arrival order enables assessing a model's performance in a reliable manner. For instance, the
    `model_selection.progressive_val_score` is a high-level method that can be used to score a
    model on a dataset. Under the hood it uses this method to determine the correct arrival order.

    Parameters:
        X_y: A stream of (features, target) tuples.
        moment: The attribute used for measuring time. If a callable is passed, then it is expected
            to take as input a `dict` of features. If `None`, then the observations are implicitely
            timestamped in the order in which they arrive. If a `str` is passed, then it will be
            used to obtain the time from the input features.
        delay: The amount of time to wait before revealing the target associated with each
            observation to the model. This value is expected to be able to sum with the `moment`
            value. For instance, if `moment` is a `datetime.date`, then `delay` is expected to be a
            `datetime.timedelta`. If a callable is passed, then it is expected to take as input a
            `dict` of features and the target. If a `str` is passed, then it will be used to access
            the relevant field from the features. If `None` is passed, then no delay will be used,
            which leads to doing standard online validation. If a scalar is passed, such an `int`
            or a `datetime.timedelta`, then the delay is constant.
        copy: If `True`, then a separate copy of the features are yielded the second time
            around. This ensures that inadvertent modifications in downstream code don't have any
            effect.

    Example:

        The arrival delay isn't usually indicated in a dataset, but it might be able to be inferred
        from the features. As an example, we'll simulate the departure and arrival time of taxi
        trips. Let's first create a time table which records the departure time and the duration of
        seconds of several taxi trips.

        >>> import datetime as dt
        >>> time_table = [
        ...     (dt.datetime(2020, 1, 1, 20,  0, 0),  900),
        ...     (dt.datetime(2020, 1, 1, 20, 10, 0), 1800),
        ...     (dt.datetime(2020, 1, 1, 20, 20, 0),  300),
        ...     (dt.datetime(2020, 1, 1, 20, 45, 0),  400),
        ...     (dt.datetime(2020, 1, 1, 20, 50, 0),  240),
        ...     (dt.datetime(2020, 1, 1, 20, 55, 0),  450)
        ... ]

        We can now create a streaming dataset where the features are the departure dates and the
        targets are the durations.

        >>> X_y = (
        ...     ({'date': date}, duration)
        ...     for date, duration in time_table
        ... )

        Now, we can use `simulate_qa` to iterate over the events in the order in which they are
        meant to occur.

        >>> delay = lambda _, y: dt.timedelta(seconds=y)

        >>> for i, x, y in simulate_qa(X_y, moment='date', delay=delay):
        ...     if y is None:
        ...         print(f'{x["date"]} - trip #{i} departs')
        ...     else:
        ...         arrival_date = x['date'] + dt.timedelta(seconds=y)
        ...         print(f'{arrival_date} - trip #{i} arrives after {y} seconds')
        2020-01-01 20:00:00 - trip #0 departs
        2020-01-01 20:10:00 - trip #1 departs
        2020-01-01 20:15:00 - trip #0 arrives after 900 seconds
        2020-01-01 20:20:00 - trip #2 departs
        2020-01-01 20:25:00 - trip #2 arrives after 300 seconds
        2020-01-01 20:40:00 - trip #1 arrives after 1800 seconds
        2020-01-01 20:45:00 - trip #3 departs
        2020-01-01 20:50:00 - trip #4 departs
        2020-01-01 20:51:40 - trip #3 arrives after 400 seconds
        2020-01-01 20:54:00 - trip #4 arrives after 240 seconds
        2020-01-01 20:55:00 - trip #5 departs
        2020-01-01 21:02:30 - trip #5 arrives after 450 seconds

        This function is extremely practical because it provides a reliable way to evaluate the
        performance of a model in a real scenario. Indeed, it allows to make predictions and
        perform model updates in exactly the same manner that would happen live. For instance, it
        is used in `model_selection.progressive_val_score`, which is a higher level function for
        evaluating models in an online manner.

    """

    # Determine how to insert mementos into the queue
    queue = lambda q, el: q.append(el)
    if callable(delay) or isinstance(delay, str):
        queue = lambda q, el: bisect.insort(q, el)

    # Coerce moment to a function
    if moment is None:
        get_moment = lambda i, _: i
    elif isinstance(moment, str):
        get_moment = lambda _, x: x[moment]
    elif callable(moment):
        get_moment = lambda _, x: moment(x)

    # Coerce delay to a function
    if delay is None:
        get_delay = lambda _, __: 0
    elif isinstance(delay, str):
        get_delay = lambda x, _: x[delay]
    elif not callable(delay):
        get_delay = lambda _, __: delay
    else:
        get_delay = delay

    mementos: typing.List[Memento] = []

    for i, (x, y) in enumerate(X_y):

        t = get_moment(i, x)
        d = get_delay(x, y)

        while mementos:

            # Get the oldest answer
            i_old, x_old, y_old, t_expire = mementos[0]

            # If the oldest answer isn't old enough then stop
            if t_expire > t:
                break

            # Reveal the duration and pop the trip from the queue
            yield i_old, x_old, y_old
            del mementos[0]

        queue(mementos, Memento(i, x, y, t + d))
        if copy:
            x = deepcopy(x)
        yield i, x, None

    for memento in mementos:
        yield memento.i, memento.x, memento.y



import collections

from creme import stats


class Shift(stats.Univariate):
    """Shifts a data stream by returning past values.

    This can be used to compute statistics over past data. For instance, if you're computing daily
    averages, then shifting by 7 will be equivalent to computing averages from a week ago.

    Parameters:
        amount: Shift amount. The `get` method will return the `t - amount` value, where `t` is the
            current moment.
        fill_value: This value will be returned by the `get` method if not enough values have been
            observed.

    Example:

        It is rare to have to use `Shift` by itself. A more common usage is to compose it with
        other statistics. This can be done via the `|` operator.

        >>> from creme import stats

        >>> stat = stats.Shift(1) | stats.Mean()

        >>> for i in range(5):
        ...     stat = stat.update(i)
        ...     print(stat.get())
        0.0
        0.0
        0.5
        1.0
        1.5

        A common usecase for using `Shift` is when computing statistics on shifted data. For
        instance, say you have a dataset which records the amount of sales for a set of shops. You
        might then have a `shop` field and a `sales` field. Let's say you want to look at the
        average amount of sales per shop. You can do this by using a `feature_extraction.Agg`. When
        you call `transform_one`, you're expecting it to return the average amount of sales,
        *without* including today's sales. You can do this by prepending an instance of
        `stats.Mean` with an instance of `stats.Shift`.

        >>> from creme import feature_extraction

        >>> agg = feature_extraction.Agg(
        ...     on='sales',
        ...     how=stats.Shift(1) | stats.Mean(),
        ...     by='shop'
        ... )

        Let's define a little example dataset.

        >>> X = iter([
        ...     {'shop': 'Ikea', 'sales': 10},
        ...     {'shop': 'Ikea', 'sales': 15},
        ...     {'shop': 'Ikea', 'sales': 20}
        ... ])

        Now let's call the `fit_one` method to update our feature extractor.

        >>> x = next(X)
        >>> agg = agg.fit_one(x)

        At this point, the average defaults to the initial value of `stats.Mean`, which is 0.

        >>> agg.transform_one(x)
        {'sales_mean_of_shift_1_by_shop': 0.0}

        We can now update our feature extractor with the next data point and check the output.

        >>> agg = agg.fit_one(next(X))
        >>> agg.transform_one(x)
        {'sales_mean_of_shift_1_by_shop': 10.0}

        >>> agg = agg.fit_one(next(X))
        >>> agg.transform_one(x)
        {'sales_mean_of_shift_1_by_shop': 12.5}

    .. tip::
        Shifting values is useful when you're calculating an average over a target value. Indeed,
        in this case it's important to shift the values in order not to introduce leakage. The
        recommended way to do this is to `feature_extraction.TargetAgg`, which already takes care
        of shifting the target values once.

    """

    def __init__(self, amount=1, fill_value=None):
        self.amount = amount
        self.fill_value = fill_value
        self.buffer = collections.deque(maxlen=self.amount + 1)

    def update(self, x):
        self.buffer.append(x)
        return self

    def get(self):
        try:
            return self.buffer[-self.amount - 1]
        except IndexError:
            return self.fill_value

    @property
    def name(self):
        return f'shift_{self.amount}'



from creme import stream

from . import base


class TrumpApproval(base.FileDataset):
    """Donald Trump approval ratings.

    This dataset was obtained by reshaping the data used by FiveThirtyEight for analyzing Donald
    Trump's approval ratings. It contains 5 features, which are approval ratings collected by
    5 polling agencies. The target is the approval rating from FiveThirtyEight's model. The goal of
    this task is to see if we can reproduce FiveThirtyEight's model.

    References:
        1. [Trump Approval Ratings](https://projects.fivethirtyeight.com/trump-approval-ratings/)

    """

    def __init__(self):
        super().__init__(
            n_samples=1001,
            n_features=6,
            task=base.REG,
            filename='trump_approval.csv.gz'
        )

    def __iter__(self):
        return stream.iter_csv(
            self.path,
            target='five_thirty_eight',
            converters={
                'ordinal_date': int,
                'gallup': float,
                'ipsos': float,
                'morning_consult': float,
                'rasmussen': float,
                'you_gov': float,
                'five_thirty_eight': float
            }
        )



from creme import linear_model
from creme import preprocessing


def test_set_params():

    obj = linear_model.LinearRegression(l2=42)
    obj.fit_one({'x': 3}, 6)

    new = obj._set_params({'l2': 21})
    assert new.l2 == 21
    assert obj.l2 == 42
    assert new.weights == {}
    assert new.weights != obj.weights


def test_set_params_pipeline():

    obj = preprocessing.StandardScaler() | linear_model.LinearRegression(l2=42)
    obj.fit_one({'x': 3}, 6)

    params = {'LinearRegression': {'l2': 21}}
    new = obj._set_params(params)
    assert new['LinearRegression'].l2 == 21
    assert obj['LinearRegression'].l2 == 42
    assert new['LinearRegression'].weights == {}
    assert new['LinearRegression'].weights != obj['LinearRegression'].weights



import collections

import numpy as np

from creme import base


class ClassificationSampler(base.Wrapper, base.Classifier):

    def __init__(self, classifier, seed=None):
        self.classifier = classifier
        self.seed = seed
        self._rng = np.random.RandomState(seed)

    @property
    def _wrapped_model(self):
        return self.classifier

    def predict_proba_one(self, x):
        return self.classifier.predict_proba_one(x)

    def predict_one(self, x):
        return self.classifier.predict_one(x)


class RandomUnderSampler(ClassificationSampler):
    """Random under-sampling.

    This is a wrapper for classifiers. It will train the provided classifier by under-sampling the
    stream of given observations so that the class distribution seen by the classifier follows
    a given desired distribution. The implementation is a discrete version of rejection sampling.

    Parameters:
        classifier
        desired_dist: The desired class distribution. The keys are the classes whilst the
            values are the desired class percentages. The values must sum up to 1.
        seed: Random seed for reproducibility.

    See the [user guide](/user-guide/imbalanced-learning) for example usage.

    References:
        1. [Under-sampling a dataset with desired ratios](https://maxhalford.github.io/blog/under-sampling-a-dataset-with-desired-ratios/)
        2. [Wikipedia article on rejection sampling](https://www.wikiwand.com/en/Rejection_sampling)

    """

    def __init__(self, classifier: base.Classifier, desired_dist: dict, seed: int = None):
        super().__init__(classifier=classifier, seed=seed)
        self.desired_dist = desired_dist
        self._actual_dist = collections.Counter()
        self._pivot = None

    def fit_one(self, x, y):

        self._actual_dist[y] += 1
        f = self.desired_dist
        g = self._actual_dist

        # Check if the pivot needs to be changed
        if y != self._pivot:
            self._pivot = max(g.keys(), key=lambda y: f[y] / g[y])
        else:
            self.classifier.fit_one(x, y)
            return self

        # Determine the sampling ratio if the class is not the pivot
        M = f[self._pivot] / g[self._pivot]  # Likelihood ratio
        ratio = f[y] / (M * g[y])

        if ratio < 1 and self._rng.random() < ratio:
            self.classifier.fit_one(x, y)

        return self


class RandomOverSampler(ClassificationSampler):
    """Random over-sampling.

    This is a wrapper for classifiers. It will train the provided classifier by over-sampling the
    stream of given observations so that the class distribution seen by the classifier follows
    a given desired distribution. The implementation is a discrete version of reverse rejection
    sampling.

    Parameters:
        classifier (base.Classifier)
        desired_dist (dict): The desired class distribution. The keys are the classes whilst the
            values are the desired class percentages. The values must sum up to 1.
        seed (int): Random seed for reproducibility.

    See :ref:`Working with imbalanced data` for example usage.

    """

    def __init__(self, classifier, desired_dist, seed=None):
        super().__init__(classifier=classifier, seed=seed)
        self.desired_dist = desired_dist
        self._actual_dist = collections.Counter()
        self._pivot = None

    def fit_one(self, x, y):

        self._actual_dist[y] += 1
        f = self.desired_dist
        g = self._actual_dist

        # Check if the pivot needs to be changed
        if y != self._pivot:
            self._pivot = max(g.keys(), key=lambda y: g[y] / f[y])
        else:
            self.classifier.fit_one(x, y)
            return self

        M = g[self._pivot] / f[self._pivot]
        rate = M * f[y] / g[y]

        for _ in range(self._rng.poisson(rate)):
            self.classifier.fit_one(x, y)

        return self


class RandomSampler(ClassificationSampler):
    """Random sampling by mixing under-sampling and over-sampling.

    This is a wrapper for classifiers. It will train the provided classifier by both under-sampling
    and over-sampling the stream of given observations so that the class distribution seen by the
    classifier follows a given desired distribution.

    Parameters:
        classifier (base.Classifier)
        desired_dist (dict): The desired class distribution. The keys are the classes whilst the
            values are the desired class percentages. The values must sum up to 1. If set to
            ``None``, then the observations will be sampled uniformly at random, which is stricly
            equivalent to using `ensemble.BaggingClassifier`.
        sampling_rate (float): The desired ratio of data to sample.
        seed (int): Random seed for reproducibility.

    See :ref:`Working with imbalanced data` for example usage.

    """

    def __init__(self, classifier, desired_dist, sampling_rate=1., seed=None):
        super().__init__(classifier=classifier, seed=seed)
        self.sampling_rate = sampling_rate
        self._actual_dist = collections.Counter()
        if desired_dist is None:
            desired_dist = self._actual_dist
        self.desired_dist = desired_dist
        self._n = 0

    def fit_one(self, x, y):

        self._actual_dist[y] += 1
        self._n += 1
        f = self.desired_dist
        g = self._actual_dist

        rate = self.sampling_rate * f[y] / (g[y] / self._n)

        for _ in range(self._rng.poisson(rate)):
            self.classifier.fit_one(x, y)

        return self



"""General tests that all estimators need to pass."""
import copy
import importlib
import inspect

import pytest

from creme import base
from creme import dummy
from creme import compat
from creme import cluster
from creme import compose
from creme import ensemble
from creme import facto
from creme import feature_extraction
from creme import feature_selection
from creme import impute
from creme import linear_model
from creme import meta
from creme import model_selection
from creme import multiclass
from creme import multioutput
from creme import naive_bayes
from creme import preprocessing
from creme import reco
from creme import sampling
from creme import stats
from creme import time_series
from creme import tree
from creme import utils
from creme.compat.sklearn import Creme2SKLBase
from creme.compat.sklearn import SKL2CremeBase


def get_all_estimators():

    ignored = (
        Creme2SKLBase,
        SKL2CremeBase,
        compose.FuncTransformer,
        compose.Pipeline,
        compose.Grouper,
        ensemble.StackingBinaryClassifier,
        facto.FFMClassifier,
        facto.FFMRegressor,
        facto.FMClassifier,
        facto.FMRegressor,
        facto.FwFMClassifier,
        facto.FwFMRegressor,
        facto.HOFMClassifier,
        facto.HOFMRegressor,
        feature_extraction.Agg,
        feature_extraction.TargetAgg,
        feature_selection.PoissonInclusion,
        impute.PreviousImputer,
        impute.StatImputer,
        linear_model.SoftmaxRegression,
        meta.PredClipper,
        meta.TransformedTargetRegressor,
        model_selection.SuccessiveHalvingClassifier,
        model_selection.SuccessiveHalvingRegressor,
        preprocessing.OneHotEncoder,
        reco.Baseline,
        reco.BiasedMF,
        reco.FunkMF,
        reco.RandomNormal,
        sampling.HardSamplingClassifier,
        sampling.HardSamplingRegressor,
        sampling.RandomOverSampler,
        sampling.RandomUnderSampler,
        sampling.RandomSampler,
        time_series.Detrender,
        time_series.GroupDetrender,
        time_series.SNARIMAX
    )

    try:
        ignored = (*ignored, compat.PyTorch2CremeRegressor)
    except AttributeError:
        pass

    def is_estimator(obj):
        return inspect.isclass(obj) and issubclass(obj, base.Estimator)

    for submodule in importlib.import_module('creme').__all__:

        if submodule == 'base':
            continue

        for _, obj in inspect.getmembers(importlib.import_module(f'creme.{submodule}'), is_estimator):

            if issubclass(obj, ignored):
                continue

            elif issubclass(obj, multioutput.RegressorChain):
                inst = obj(model=linear_model.LinearRegression())

            elif issubclass(obj, multioutput.ClassifierChain):
                inst = obj(model=linear_model.LogisticRegression())

            elif issubclass(obj, dummy.StatisticRegressor):
                inst = obj(statistic=stats.Mean())

            elif issubclass(obj, tree.RandomForestClassifier):
                inst = obj()

            elif issubclass(obj, ensemble.BaggingClassifier):
                inst = obj(linear_model.LogisticRegression())

            elif issubclass(obj, ensemble.BaggingRegressor):
                inst = obj(linear_model.LinearRegression())

            elif issubclass(obj, ensemble.AdaBoostClassifier):
                inst = obj(linear_model.LogisticRegression())

            elif issubclass(obj, ensemble.HedgeRegressor):
                inst = obj([
                    preprocessing.StandardScaler() | linear_model.LinearRegression(intercept_lr=.1),
                    preprocessing.StandardScaler() | linear_model.PARegressor(),
                ])

            elif issubclass(obj, feature_selection.SelectKBest):
                inst = obj(similarity=stats.PearsonCorr())

            elif issubclass(obj, linear_model.LinearRegression):
                inst = preprocessing.StandardScaler() | obj(intercept_lr=.1)

            elif issubclass(obj, linear_model.PARegressor):
                inst = preprocessing.StandardScaler() | obj()

            elif issubclass(obj, multiclass.OneVsRestClassifier):
                inst = obj(classifier=linear_model.LogisticRegression())

            elif issubclass(obj, multiclass.OneVsOneClassifier):
                inst = obj(classifier=linear_model.LogisticRegression())

            elif issubclass(obj, multiclass.OutputCodeClassifier):
                inst = obj(classifier=linear_model.LogisticRegression(), code_size=10)

            else:
                inst = obj()

            yield inst


@pytest.mark.parametrize('estimator, check', [
    pytest.param(
        estimator,
        check,
        id=f'{estimator}:{check.__name__}'
    )
    for estimator in list(get_all_estimators()) + [
        feature_extraction.TFIDF(),
        linear_model.LogisticRegression(),
        preprocessing.StandardScaler() | linear_model.LinearRegression(),
        preprocessing.StandardScaler() | linear_model.PAClassifier(),
        preprocessing.StandardScaler() | multiclass.OneVsRestClassifier(linear_model.LogisticRegression()),
        preprocessing.StandardScaler() | multiclass.OneVsRestClassifier(linear_model.PAClassifier()),
        naive_bayes.GaussianNB(),
        preprocessing.StandardScaler(),
        cluster.KMeans(n_clusters=5, seed=42),
        preprocessing.MinMaxScaler(),
        preprocessing.MinMaxScaler() + preprocessing.StandardScaler(),
        preprocessing.PolynomialExtender(),
        feature_selection.VarianceThreshold(),
        feature_selection.SelectKBest(similarity=stats.PearsonCorr())
    ]
    for check in utils.estimator_checks.yield_checks(estimator)
])
def test_check_estimator(estimator, check):
    check(copy.deepcopy(estimator))



import collections
import functools
import random

from creme import base
from creme import utils


__all__ = ['KMeans']


class KMeans(base.Clusterer):
    """Incremental k-means.

    The most common way to implement batch k-means is to use Lloyd's algorithm, which consists in
    assigning all the data points to a set of cluster centers and then moving the centers
    accordingly. This requires multiple passes over the data and thus isn't applicable in a
    streaming setting.

    In this implementation we start by finding the cluster that is closest to the current
    observation. We then move the cluster's central position towards the new observation. The
    `halflife` parameter determines by how much to move the cluster toward the new observation.
    You will get better results if you scale your data appropriately.

    Parameters:
        n_clusters: Maximum number of clusters to assign.
        halflife: Amount by which to move the cluster centers, a reasonable value if
            between 0 and 1.
        mu: Mean of the normal distribution used to instantiate cluster positions.
        sigma: Standard deviation of the normal distribution used to instantiate cluster positions.
        p: Power parameter for the Minkowski metric. When `p=1`, this corresponds to the
            Manhattan distance, while `p=2` corresponds to the Euclidean distance.
        seed: Random seed used for generating initial centroid positions.

    Attributes:
        centers (dict): Central positions of each cluster.

    Example:

        In the following example the cluster assignments are exactly the same as when using
        `sklearn`'s batch implementation. However changing the `halflife` parameter will
        produce different outputs.

        >>> from creme import cluster
        >>> from creme import stream

        >>> X = [
        ...     [1, 2],
        ...     [1, 4],
        ...     [1, 0],
        ...     [4, 2],
        ...     [4, 4],
        ...     [4, 0]
        ... ]

        >>> k_means = cluster.KMeans(n_clusters=2, halflife=0.4, sigma=3, seed=0)

        >>> for i, (x, _) in enumerate(stream.iter_array(X)):
        ...     k_means = k_means.fit_one(x)
        ...     print(f'{X[i]} is assigned to cluster {k_means.predict_one(x)}')
        [1, 2] is assigned to cluster 1
        [1, 4] is assigned to cluster 1
        [1, 0] is assigned to cluster 0
        [4, 2] is assigned to cluster 0
        [4, 4] is assigned to cluster 0
        [4, 0] is assigned to cluster 0

        >>> k_means.predict_one({0: 0, 1: 0})
        1

        >>> k_means.predict_one({0: 4, 1: 4})
        0

    References:
        1. [Sequential k-Means Clustering](http://www.cs.princeton.edu/courses/archive/fall08/cos436/Duda/C/sk_means.htm)
        2. [Sculley, D., 2010, April. Web-scale k-means clustering. In Proceedings of the 19th international conference on World wide web (pp. 1177-1178](https://www.eecs.tufts.edu/~dsculley/papers/fastkmeans.pdf)

    """

    def __init__(self, n_clusters=5, halflife=0.5, mu=0, sigma=1, p=2, seed: int = None):
        self.n_clusters = n_clusters
        self.halflife = halflife
        self.mu = mu
        self.sigma = sigma
        self.p = p
        self.seed = seed
        self._rng = random.Random(seed)
        rand_gauss = functools.partial(self._rng.gauss, self.mu, self.sigma)
        self.centers = {i: collections.defaultdict(rand_gauss) for i in range(n_clusters)}  # type: ignore

    def fit_predict_one(self, x):
        """Equivalent to `k_means.fit_one(x).predict_one(x)`, but faster."""

        # Find the cluster with the closest center
        closest = self.predict_one(x)

        # Move the cluster's center
        for i, xi in x.items():
            self.centers[closest][i] += self.halflife * (xi - self.centers[closest][i])

        return closest

    def fit_one(self, x):
        self.fit_predict_one(x)
        return self

    def predict_one(self, x):

        def get_distance(c):
            return utils.math.minkowski_distance(a=self.centers[c], b=x, p=self.p)

        return min(self.centers, key=get_distance)



import math

from creme import base


__all__ = ['BoxCoxRegressor', 'TransformedTargetRegressor']


class TransformedTargetRegressor(base.Regressor, base.Wrapper):
    """Modifies the target before training.

    The user is expected to check that `func` and `inverse_func` are coherent with each other.

    Parameters:
        regressor: regressor model applied for training.
        func: A function modifying the target before training.
        inverse_func: A function to return to the target's original space.

    Example:

        >>> import math
        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import meta
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import preprocessing

        >>> X_y = datasets.TrumpApproval()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     meta.TransformedTargetRegressor(
        ...         regressor=linear_model.LinearRegression(intercept_lr=0.15),
        ...         func=math.log,
        ...         inverse_func=math.exp
        ...     )
        ... )
        >>> metric = metrics.MSE()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MSE: 8.759624

    """

    def __init__(self, regressor: base.Regressor, func: callable, inverse_func: callable):
        self.regressor = regressor
        self.func = func
        self.inverse_func = inverse_func

    @property
    def _wrapped_model(self):
        return self.regressor

    def fit_one(self, x, y):
        self.regressor.fit_one(x, self.func(y))
        return self

    def predict_one(self, x):
        y_pred = self.regressor.predict_one(x)
        return self.inverse_func(y_pred)


class BoxCoxRegressor(TransformedTargetRegressor):
    """Applies the Box-Cox transform to the target before training.

    Box-Cox transform is useful when the target variable is heteroscedastic (i.e. there are
    sub-populations that have different variabilities from others) allowing to transform it towards
    normality.

    The `power` parameter is denoted λ in the litterature. If `power` is equal to 0 than the
    Box-Cox transform will be equivalent to a log transform.

    Parameters:
        regressor: regressor model applied for training.
        power: power value to do the transformation.

    Example:

        >>> import math
        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import meta
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import preprocessing

        >>> X_y = datasets.TrumpApproval()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     meta.BoxCoxRegressor(
        ...         regressor=linear_model.LinearRegression(intercept_lr=0.2),
        ...         power=0.05
        ...     )
        ... )
        >>> metric = metrics.MSE()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MSE: 5.898196

    """

    def __init__(self, regressor: base.Regressor, power=1.):
        super().__init__(
            regressor=regressor,
            func=(lambda y: (y ** power - 1) / power) if power > 0 else math.log,
            inverse_func=(lambda y: (power * y + 1) ** (1 / power)) if power > 0 else math.exp
        )



from creme import stream

from . import base


class Music(base.RemoteDataset):
    """Bike sharing station information from the city of Toulouse.

    The goal is to predict to which kinds of moods a song pertains to.

    References:
        1. [Read, J., Reutemann, P., Pfahringer, B. and Holmes, G., 2016. MEKA: a multi-label/multi-target extension to WEKA. The Journal of Machine Learning Research, 17(1), pp.667-671.](http://www.jmlr.org/papers/v17/12-164.html)

    """

    def __init__(self):
        super().__init__(
            n_samples=593,
            n_features=72,
            task=base.MO_BINARY_CLF,
            url='https://raw.githubusercontent.com/scikit-multiflow/streaming-datasets/master/music.csv',
            size=378980,
            unpack=False
        )

    def _iter(self):
        return stream.iter_csv(
            self.path,
            target=[
                'amazed-suprised',
                'happy-pleased',
                'relaxing-clam',
                'quiet-still',
                'sad-lonely',
                'angry-aggresive'
            ],
            converters={
                'amazed-suprised': lambda x: x == '1',
                'happy-pleased': lambda x: x == '1',
                'relaxing-clam': lambda x: x == '1',
                'quiet-still': lambda x: x == '1',
                'sad-lonely': lambda x: x == '1',
                'angry-aggresive': lambda x: x == '1',
                'Mean_Acc1298_Mean_Mem40_Centroid': float,
                'Mean_Acc1298_Mean_Mem40_Rolloff': float,
                'Mean_Acc1298_Mean_Mem40_Flux': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_0': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_1': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_2': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_3': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_4': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_5': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_6': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_7': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_8': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_9': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_10': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_11': float,
                'Mean_Acc1298_Mean_Mem40_MFCC_12': float,
                'Mean_Acc1298_Std_Mem40_Centroid': float,
                'Mean_Acc1298_Std_Mem40_Rolloff': float,
                'Mean_Acc1298_Std_Mem40_Flux': float,
                'Mean_Acc1298_Std_Mem40_MFCC_0': float,
                'Mean_Acc1298_Std_Mem40_MFCC_1': float,
                'Mean_Acc1298_Std_Mem40_MFCC_2': float,
                'Mean_Acc1298_Std_Mem40_MFCC_3': float,
                'Mean_Acc1298_Std_Mem40_MFCC_4': float,
                'Mean_Acc1298_Std_Mem40_MFCC_5': float,
                'Mean_Acc1298_Std_Mem40_MFCC_6': float,
                'Mean_Acc1298_Std_Mem40_MFCC_7': float,
                'Mean_Acc1298_Std_Mem40_MFCC_8': float,
                'Mean_Acc1298_Std_Mem40_MFCC_9': float,
                'Mean_Acc1298_Std_Mem40_MFCC_10': float,
                'Mean_Acc1298_Std_Mem40_MFCC_11': float,
                'Mean_Acc1298_Std_Mem40_MFCC_12': float,
                'Std_Acc1298_Mean_Mem40_Centroid': float,
                'Std_Acc1298_Mean_Mem40_Rolloff': float,
                'Std_Acc1298_Mean_Mem40_Flux': float,
                'Std_Acc1298_Mean_Mem40_MFCC_0': float,
                'Std_Acc1298_Mean_Mem40_MFCC_1': float,
                'Std_Acc1298_Mean_Mem40_MFCC_2': float,
                'Std_Acc1298_Mean_Mem40_MFCC_3': float,
                'Std_Acc1298_Mean_Mem40_MFCC_4': float,
                'Std_Acc1298_Mean_Mem40_MFCC_5': float,
                'Std_Acc1298_Mean_Mem40_MFCC_6': float,
                'Std_Acc1298_Mean_Mem40_MFCC_7': float,
                'Std_Acc1298_Mean_Mem40_MFCC_8': float,
                'Std_Acc1298_Mean_Mem40_MFCC_9': float,
                'Std_Acc1298_Mean_Mem40_MFCC_10': float,
                'Std_Acc1298_Mean_Mem40_MFCC_11': float,
                'Std_Acc1298_Mean_Mem40_MFCC_12': float,
                'Std_Acc1298_Std_Mem40_Centroid': float,
                'Std_Acc1298_Std_Mem40_Rolloff': float,
                'Std_Acc1298_Std_Mem40_Flux': float,
                'Std_Acc1298_Std_Mem40_MFCC_0': float,
                'Std_Acc1298_Std_Mem40_MFCC_1': float,
                'Std_Acc1298_Std_Mem40_MFCC_2': float,
                'Std_Acc1298_Std_Mem40_MFCC_3': float,
                'Std_Acc1298_Std_Mem40_MFCC_4': float,
                'Std_Acc1298_Std_Mem40_MFCC_5': float,
                'Std_Acc1298_Std_Mem40_MFCC_6': float,
                'Std_Acc1298_Std_Mem40_MFCC_7': float,
                'Std_Acc1298_Std_Mem40_MFCC_8': float,
                'Std_Acc1298_Std_Mem40_MFCC_9': float,
                'Std_Acc1298_Std_Mem40_MFCC_10': float,
                'Std_Acc1298_Std_Mem40_MFCC_11': float,
                'Std_Acc1298_Std_Mem40_MFCC_12': float,
                'BH_LowPeakAmp': float,
                'BH_LowPeakBPM': int,
                'BH_HighPeakAmp': float,
                'BH_HighPeakBPM': int,
                'BH_HighLowRatio': int,
                'BHSUM1': float,
                'BHSUM2': float,
                'BHSUM3': float
            }
        )



import pytest

from creme import compose
from creme import linear_model
from creme import model_selection
from creme import optim
from creme import preprocessing
from creme import tree


@pytest.mark.parametrize('param_grid, count', [
    (
        {
            'LinearRegression': {
                'optimizer': [
                    (optim.SGD, {'lr': [1, 2]}),
                    (optim.Adam, {
                        'beta_1': [0.1, 0.01, 0.001],
                        'lr': [0.1, 0.01, 0.001, 0.0001]
                    })
                ]
            }
        },
        2 + 3 * 4
    ),
    (
        {
            'Scaler': [
                preprocessing.MinMaxScaler(),
                preprocessing.MaxAbsScaler(),
                preprocessing.StandardScaler()
            ],
            'LinearRegression': {
                'optimizer': {
                    'lr': [1e-1, 1e-2, 1e-3]
                }
            }
        },
        3 * 3
    )

])
def test_expand_param_grid_count(param_grid, count):
    assert sum(1 for _ in model_selection.expand_param_grid(param_grid)) == count


def test_decision_tree_max_depth():

    model = tree.DecisionTreeClassifier()

    max_depths = [1, 2, 3, 4, 5, 6]
    param_grid = model_selection.expand_param_grid({
        'max_depth': max_depths
    })

    models = [model._set_params(params) for params in param_grid]

    for model, max_depth in zip(models, max_depths):
        assert model.max_depth == max_depth



"""Weight initializers."""
import abc

import numpy as np


__all__ = [
    'Constant',
    'Normal',
    'Zeros'
]


class Initializer(abc.ABC):
    """An initializer is used to set initial weights in a model."""

    def __str__(self):
        return self.__class__.__name__

    @abc.abstractmethod
    def __call__(self, shape=1):
        """Returns a fresh set of weights.

        Parameters:
            shape: Indicates how many weights to return. If `1`, then a single scalar value will
                be returned.

        """


class Constant(Initializer):
    """Constant initializer which always returns the same value.

    Parameters:
        value

    Example:

        >>> from creme import optim

        >>> init = optim.initializers.Constant(value=3.14)

        >>> init(shape=1)
        3.14

        >>> init(shape=2)
        array([3.14, 3.14])

    """

    def __init__(self, value: float):
        self.value = value

    def __call__(self, shape=1):
        return np.full(shape, self.value) if shape != 1 else self.value


class Zeros(Constant):
    """Constant initializer which always returns zeros.

    Example:

        >>> from creme import optim

        >>> init = optim.initializers.Zeros()

        >>> init(shape=1)
        0.0

        >>> init(shape=2)
        array([0., 0.])

    """

    def __init__(self):
        super().__init__(value=0.)


class Normal(Initializer):
    """Random normal initializer which simulate a normal distribution with specified parameters.

    Parameters:
        mu: The mean of the normal distribution
        sigma: The standard deviation of the normal distribution

    Example:

        >>> from creme import optim

        >>> init = optim.initializers.Normal(mu=0, sigma=1, seed=42)

        >>> init(shape=1)
        0.496714

        >>> init(shape=2)
        array([-0.1382643 ,  0.64768854])

    """

    def __init__(self, mu=0., sigma=1., seed=None):
        self.mu = mu
        self.sigma = sigma
        self.seed = seed
        self._rng = np.random.RandomState(seed)

    def __call__(self, shape=1):
        weights = self._rng.normal(loc=self.mu, scale=self.sigma, size=shape)
        if shape == 1:
            return weights[0]
        return weights



"""Recommender systems."""
from .baseline import Baseline
from .biased_mf import BiasedMF
from .funk_mf import FunkMF
from .normal import RandomNormal


__all__ = ['Baseline', 'BiasedMF', 'FunkMF', 'RandomNormal']

try:
    from .surprise import SurpriseWrapper
    __all__ += ['SurpriseWrapper']
except ImportError:
    pass



import collections
import functools
import typing

from creme import base
from creme import stats


class Detrender(base.Regressor, base.Wrapper):
    """A linear detrender which centers the target in zero.

    At each `fit_one` step, the current mean of `y` is substracted from `y` before being fed
    to the provided regression model. During the `predict_one` step, the current mean is added
    to the prediction of the regression model.

    Parameters:
        regressor
        window_size: Window size used for calculating the rolling mean. If `None`, then a mean over
            the whole target data will instead be used.

    """

    def __init__(self, regressor: base.Regressor, window_size: int = None):
        self.regressor = regressor
        self.mean = stats.Mean() if window_size is None else stats.RollingMean(window_size)

    @property
    def _wrapped_model(self):
        return self.regressor

    def fit_one(self, x, y):
        self.regressor.fit_one(x, y - self.mean.get())
        self.mean.update(y)
        return self

    def predict_one(self, x):
        return self.regressor.predict_one(x) + self.mean.get()


class GroupDetrender(base.Regressor, base.Wrapper):
    """Removes the trend of the target inside each group.

    Parameters:
        regressor
        by
        window_size: Window size used for calculating each rolling mean. If `None`, then a
            mean over the whole target data will instead be used.

    """

    def __init__(self, regressor: base.Regressor, by: str, window_size: int = None):
        self.regressor = regressor
        self.by = by
        self.means: typing.DefaultDict[typing.Any, stats.Univariate] = collections.defaultdict(
            stats.Mean if window_size is None else
            functools.partial(stats.RollingMean, window_size)
        )

    @property
    def _wrapped_model(self):
        return self.regressor

    def fit_one(self, x, y):
        key = x[self.by]
        self.regressor.fit_one(x, y - self.means[key].get())
        self.means[key].update(y)
        return self

    def predict_one(self, x):
        key = x[self.by]
        return self.regressor.predict_one(x) + self.means[key].get()



import abc
import numbers
import typing

from creme import utils

from . import schedulers


class Optimizer(abc.ABC):
    """Optimizer interface.

    Every optimizer inherits from this base interface.

    Parameters:
        lr

    Attributes:
        learning_rate (float): Returns the current learning rate value.

    """

    def __init__(self, lr: typing.Union[schedulers.Scheduler, float]):
        if isinstance(lr, numbers.Number):
            lr = schedulers.Constant(lr)
        self.lr = lr
        self.n_iterations = 0

    @property
    def learning_rate(self) -> float:
        return self.lr.get(self.n_iterations)

    def update_before_pred(self, w: dict) -> dict:
        """Updates a weight vector before a prediction is made.

        Parameters:
            w (dict): A dictionary of weight parameters. The weights are modified in-place.

        Returns:
            The updated weights.

        """
        return w

    @abc.abstractmethod
    def _update_after_pred(self, w: dict, g: dict) -> dict:
        """Updates a weight vector given a gradient."""
        raise NotImplementedError

    def update_after_pred(self, w: dict, g: dict) -> dict:
        """Updates a weight vector given a gradient.

        Parameters:
            w (dict): A dictionary of weight parameters. The weights are modified in-place.
            g (dict): A dictionary of gradients.

        Returns:
            The updated weights.

        """

        # Update the weights
        w = self._update_after_pred(w, g)

        # Update the iteration counter
        self.n_iterations += 1

        return w

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return f'{self.__class__.__name__}({vars(self)})'



import abc
import collections
import functools
import operator

from creme import base
from creme import proba
from creme import utils


def decimal_range(start, stop, num):
    """

    Example:

        >>> for x in decimal_range(0, 1, 4):
        ...     print(x)
        0.2
        0.4
        0.6
        0.8

    """
    step = (stop - start) / (num + 1)

    for _ in range(num):
        start += step
        yield start


class Op:

    __slots__ = 'symbol', 'func'

    def __init__(self, symbol, func):
        self.symbol = symbol
        self.func = func

    def __call__(self, a, b):
        return self.func(a, b)

    def __repr__(self):
        return self.symbol


LT = Op('<', operator.lt)
EQ = Op('=', operator.eq)


class SplitEnum(abc.ABC):

    @abc.abstractmethod
    def update(self, x, y):
        """Updates the sufficient statistics used for evaluting splits."""

    @abc.abstractmethod
    def enumerate_splits(self, target_dist):
        """Yields candidate split points and associated operators."""


class HistSplitEnum(SplitEnum):
    """Split enumerator for classification and numerical attributes."""

    def __init__(self, n_bins, n_splits):
        self.hists = collections.defaultdict(functools.partial(utils.Histogram, max_bins=n_bins))
        self.n_splits = n_splits

    def update(self, x: float, y: base.typing.ClfTarget):
        self.hists[y].update(x)
        return self

    def enumerate_splits(self, target_dist: proba.Multinomial):

        low = min(h[0].right for h in self.hists.values())
        high = min(h[-1].right for h in self.hists.values())

        # If only one single value has been observed, then no split can be proposed
        if low >= high:
            return
            yield  # not a typo

        n_thresholds = min(
            self.n_splits,
            max(map(len, self.hists.values())) - 1
        )

        thresholds = list(decimal_range(start=low, stop=high, num=n_thresholds))
        cdfs = {y: hist.iter_cdf(thresholds) for y, hist in self.hists.items()}

        for at in thresholds:

            l_dist = {}
            r_dist = {}

            for y in target_dist:
                p_xy = next(cdfs[y]) if y in cdfs else 0.  # P(x < t | y)
                p_y = target_dist.pmf(y)  # P(y)
                l_dist[y] = target_dist.n_samples * p_y * p_xy  # P(y | x < t)
                r_dist[y] = target_dist.n_samples * p_y * (1 - p_xy)  # P(y | x >= t)

            yield LT, at, proba.Multinomial(l_dist), proba.Multinomial(r_dist)


class CategoricalSplitEnum(SplitEnum):
    """Split enumerator for classification and categorical attributes."""

    def __init__(self):
        self.P_xy = collections.defaultdict(proba.Multinomial)

    def update(self, x: str, y: base.typing.ClfTarget):
        self.P_xy[y].update(x)
        return self

    def enumerate_splits(self, target_dist: proba.Multinomial):

        categories = set(*(p_x.keys() for p_x in self.P_xy.values()))

        # There has to be at least two categories for a split to be possible
        if len(categories) < 2:
            return
            yield  # not a typo

        for cat in categories:

            l_dist = {}
            r_dist = {}

            for y in target_dist:
                p_xy = self.P_xy[y].pmf(cat)  # P(cat | y)
                p_y = target_dist.pmf(y)  # P(y)
                l_dist[y] = target_dist.n_samples * p_y * p_xy  # P(y | cat)
                r_dist[y] = target_dist.n_samples * p_y * (1. - p_xy)  # P(y | !cat)

            yield EQ, cat, proba.Multinomial(l_dist), proba.Multinomial(r_dist)



import bisect
import collections
import typing


class Window:
    """Running window data structure.

    This is just a convenience layer on top of a `collections.deque`. The only reason this exists
    is that deepcopying a class which inherits from `collections.deque` seems to bug out when the
    class has a parameter with no default value.

    Parameters:
        size: Size of the rolling window.

    Example:

        >>> from creme import utils

        >>> window = utils.Window(size=2)

        >>> for x in [1, 2, 3, 4, 5, 6]:
        ...     print(window.append(x))
        [1]
        [1, 2]
        [2, 3]
        [3, 4]
        [4, 5]
        [5, 6]

    """

    def __init__(self, size: int):
        self.values: typing.Deque[typing.Any] = collections.deque(maxlen=size)

    @property
    def size(self):
        return self.values.maxlen

    def __repr__(self):
        return str(list(self.values))

    def __len__(self):
        return len(self.values)

    def __getitem__(self, idx):
        return self.values[idx]

    def __setitem__(self, idx, val):
        self.values[idx] = val
        return self

    def extend(self, values):
        self.values.extend(values)
        return self

    def append(self, x):
        self.values.append(x)
        return self


class SortedWindow(collections.UserList):
    """Sorted running window data structure.

    Parameters:
        size: size of the window to compute the rolling quantile.

    Example:

        >>> from creme import utils

        >>> window = utils.SortedWindow(size=3)

        >>> for i in reversed(range(9)):
        ...     print(window.append(i))
        [8]
        [7, 8]
        [6, 7, 8]
        [5, 6, 7]
        [4, 5, 6]
        [3, 4, 5]
        [2, 3, 4]
        [1, 2, 3]
        [0, 1, 2]

    References:
       1. [Left sorted inserts in Python](https://stackoverflow.com/questions/8024571/insert-an-item-into-sorted-list-in-python)

    """

    def __init__(self, size: int):
        super().__init__()
        self.unsorted_window = Window(size)

    @property
    def size(self):
        return self.unsorted_window.size

    def append(self, x):

        if len(self) >= self.size:
            self.remove(self.unsorted_window[0])

        bisect.insort_left(self, x)
        self.unsorted_window.append(x)

        return self



"""Datasets."""
from . import synth
from .airline_passengers import AirlinePassengers
from .bananas import Bananas
from .chick_weights import ChickWeights
from .credit_card import CreditCard
from .elec2 import Elec2
from .higgs import Higgs
from .http import HTTP
from .insects import Insects
from .music import Music
from .phishing import Phishing
from .movielens100k import MovieLens100K
from .restaurants import Restaurants
from .segment import ImageSegments
from .sms_spam import SMSSpam
from .smtp import SMTP
from .bikes import Bikes
from .taxis import Taxis
from .trec07 import TREC07
from .trump_approval import TrumpApproval
from .malicious_url import MaliciousURL


__all__ = [
    'AirlinePassengers',
    'Bananas',
    'Bikes',
    'ChickWeights',
    'CreditCard',
    'Elec2',
    'Higgs',
    'HTTP',
    'ImageSegments',
    'Insects',
    'MaliciousURL',
    'MovieLens100K',
    'Music',
    'Phishing',
    'Restaurants',
    'SMSSpam',
    'SMTP',
    'synth',
    'Taxis',
    'TREC07',
    'TrumpApproval'
]



import bisect
import datetime as dt
import typing

from . import base


__all__ = ['TimeRolling']


class TimeRolling(base.WrapperMetric):
    """Wrapper for computing metrics over a period of time.

    Parameters:
        metric: A metric.
        period: A period of time.

    Example:

        >>> import datetime as dt
        >>> from creme import metrics

        >>> y_true = [3, -0.5, 2, 7]
        >>> y_pred = [2.5, 0.0, 2, 9]
        >>> days = [1, 2, 3, 4]

        >>> metric = metrics.TimeRolling(metrics.MAE(), period=dt.timedelta(days=2))

        >>> for yt, yp, day in zip(y_true, y_pred, days):
        ...     t = dt.datetime(2019, 1, day)
        ...     print(metric.update(yt, yp, t))
        MAE: 0.5
        MAE: 0.5
        MAE: 0.25
        MAE: 1.

    """

    def __init__(self, metric: base.Metric, period: dt.timedelta):
        self._metric = metric
        self.period = period
        self._events: typing.List[typing.Tuple[dt.datetime, typing.Any, typing.Any]] = []
        self._latest = dt.datetime(1, 1, 1)

    @property
    def metric(self):
        return self._metric

    def update(self, y_true, y_pred, t):
        self.metric.update(y_true, y_pred)
        bisect.insort_left(self._events, (t, y_true, y_pred))

        # There will only be events to revert if the new event if younger than the previously seen
        # youngest event
        if t > self._latest:
            self._latest = t

            i = 0
            for ti, yt, yp in self._events:
                if ti > t - self.period:
                    break
                self.metric.revert(yt, yp)
                i += 1

            # Remove expired events
            if i > 0:
                self._events = self._events[i:]

        return self



import math

from . import base
from . import confusion


__all__ = ['MCC']


class MCC(base.BinaryMetric):
    """Matthews correlation coefficient.

    Example:

        >>> from creme import metrics

        >>> y_true = [True, True, True, False]
        >>> y_pred = [True, False, True, True]

        >>> mcc = metrics.MCC()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     mcc = mcc.update(yt, yp)

        >>> mcc
        MCC: -0.333333

    References:
        1. [Wikipedia article](https://www.wikiwand.com/en/Matthews_correlation_coefficient)

    """

    def __init__(self):
        self.cm = confusion.ConfusionMatrix()

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def update(self, y_true, y_pred, sample_weight=1.):
        self.cm.update(y_true=y_true, y_pred=y_pred, sample_weight=sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.cm.revert(y_true=y_true, y_pred=y_pred, sample_weight=sample_weight)
        return self

    def get(self):
        tp = self.cm.counts.get(True, {}).get(True, 0)
        tn = self.cm.counts.get(False, {}).get(False, 0)
        fp = self.cm.counts.get(False, {}).get(True, 0)
        fn = self.cm.counts.get(True, {}).get(False, 0)

        n = (tp + tn + fp + fn) or 1
        s = (tp + fn) / n
        p = (tp + fp) / n

        try:
            return (tp / n - s * p) / math.sqrt(p * s * (1 - s) * (1 - p))
        except ZeroDivisionError:
            return 0.



from creme import stream

from . import base


class CreditCard(base.RemoteDataset):
    """Credit card frauds.

    The datasets contains transactions made by credit cards in September 2013 by european
    cardholders. This dataset presents transactions that occurred in two days, where we have 492
    frauds out of 284,807 transactions. The dataset is highly unbalanced, the positive class
    (frauds) account for 0.172% of all transactions.

    It contains only numerical input variables which are the result of a PCA transformation.
    Unfortunately, due to confidentiality issues, we cannot provide the original features and more
    background information about the data. Features V1, V2, ... V28 are the principal components
    obtained with PCA, the only features which have not been transformed with PCA are 'Time' and
    'Amount'. Feature 'Time' contains the seconds elapsed between each transaction and the first
    transaction in the dataset. The feature 'Amount' is the transaction Amount, this feature can be
    used for example-dependant cost-senstive learning. Feature 'Class' is the response variable and
    it takes value 1 in case of fraud and 0 otherwise.

    References:
        1. Andrea Dal Pozzolo, Olivier Caelen, Reid A. Johnson and Gianluca Bontempi. Calibrating Probability with Undersampling for Unbalanced Classification. In Symposium on Computational Intelligence and Data Mining (CIDM), IEEE, 2015
        2. Dal Pozzolo, Andrea; Caelen, Olivier; Le Borgne, Yann-Ael; Waterschoot, Serge; Bontempi, Gianluca. Learned lessons in credit card fraud detection from a practitioner perspective, Expert systems with applications,41,10,4915-4928,2014, Pergamon
        3. Dal Pozzolo, Andrea; Boracchi, Giacomo; Caelen, Olivier; Alippi, Cesare; Bontempi, Gianluca. Credit card fraud detection: a realistic modeling and a novel learning strategy, IEEE transactions on neural networks and learning systems,29,8,3784-3797,2018,IEEE
        4. Dal Pozzolo, Andrea Adaptive Machine learning for credit card fraud detection ULB MLG PhD thesis (supervised by G. Bontempi)
        5. Carcillo, Fabrizio; Dal Pozzolo, Andrea; Le Borgne, Yann-Ael; Caelen, Olivier; Mazzer, Yannis; Bontempi, Gianluca. Scarff: a scalable framework for streaming credit card fraud detection with Spark, Information fusion,41, 182-194,2018,Elsevier
        6. Carcillo, Fabrizio; Le Borgne, Yann-Ael; Caelen, Olivier; Bontempi, Gianluca. Streaming active learning strategies for real-life credit card fraud detection: assessment and visualization, International Journal of Data Science and Analytics, 5,4,285-300,2018,Springer International Publishing
        7. Bertrand Lebichot, Yann-Ael Le Borgne, Liyun He, Frederic Oble, Gianluca Bontempi Deep-Learning Domain Adaptation Techniques for Credit Cards Fraud Detection, INNSBDDL 2019: Recent Advances in Big Data and Deep Learning, pp 78-88, 2019
        8. Fabrizio Carcillo, Yann-Ael Le Borgne, Olivier Caelen, Frederic Oble, Gianluca Bontempi Combining Unsupervised and Supervised Learning in Credit Card Fraud Detection Information Sciences, 2019

    """

    def __init__(self):
        super().__init__(
            n_samples=150828752,
            n_features=30,
            task=base.BINARY_CLF,
            url='https://maxhalford.github.io/files/datasets/creditcardfraud.zip',
            size=150828752,
            filename='creditcard.csv'
        )

    def _iter(self):

        converters = {f'V{i}': float for i in range(1, 29)}
        converters['Class'] = int
        converters['Time'] = float
        converters['Amount'] = float

        return stream.iter_csv(self.path, target='Class', converters=converters)



from creme import stream

from . import base


class Bikes(base.RemoteDataset):
    """Bike sharing station information from the city of Toulouse.

    The goal is to predict the number of bikes in 5 different bike stations from the city of
    Toulouse.

    References:
        1. [A short introduction and conclusion to the OpenBikes 2016 Challenge](https://maxhalford.github.io/blog/a-short-introduction-and-conclusion-to-the-openbikes-2016-challenge/)

    """

    def __init__(self):
        super().__init__(
            url='https://maxhalford.github.io/files/datasets/toulouse_bikes.zip',
            size=13125015,
            n_samples=182_470,
            n_features=8,
            task=base.REG,
            filename='toulouse_bikes.csv'
        )

    def _iter(self):
        return stream.iter_csv(
            self.path,
            target='bikes',
            converters={
                'clouds': int,
                'humidity': int,
                'pressure': float,
                'temperature': float,
                'wind': float,
                'bikes': int
            },
            parse_dates={'moment': '%Y-%m-%d %H:%M:%S'}
        )



import collections
import copy
import functools
import itertools
import random

from creme import base


__all__ = ['OutputCodeClassifier']


def l1_dist(a, b):
    return sum(abs(ai - bi) for ai, bi in zip(a, b))


class OutputCodeClassifier(base.Wrapper, base.MultiClassifier):
    """Output-code multiclass strategy.

    This also referred to as "error-correcting output codes".

    This class allows to learn a multi-class classification problem with a binary classifier. Each
    class is converted to a code of 0s and 1s. The length of the code is called  the code size. A
    copy of the classifier made for code. The codes associated with the classes are stored in a
    code book.

    When a new sample arrives, the label's code is retrieved from the code book. Then, each
    classifier is trained on the relevant part of code, which is either a 0 or a 1.

    For predicting, each classifier outputs a probability. These are then compared to each code in
    the code book, and the label which is the "closest" is chosen as the most likely class.
    Closeness is determined in terms of Manhattan distance.

    One specificity of online learning is that we don't how many classes there are initially.
    Therefore, a random procedure generates random codes on the fly whenever a previously unseed
    label appears.

    Parameters:
        classifier:  A binary classifier, although a multi-class classifier will work too.
        code_size: The code size, which dictates how many copies of the provided classifiers to
            train. Must be strictly positive.
        seed: A random seed number that can be set for reproducibility.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import multiclass
        >>> from creme import preprocessing

        >>> X_y = datasets.ImageSegments()

        >>> scaler = preprocessing.StandardScaler()
        >>> ooc = OutputCodeClassifier(
        ...     classifier=linear_model.LogisticRegression(),
        ...     code_size=10,
        ...     seed=24
        ... )
        >>> model = scaler | ooc

        >>> metric = metrics.MacroF1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MacroF1: 0.797119

    References:
        1. [Dietterich, T.G. and Bakiri, G., 1994. Solving multiclass learning problems via error-correcting output codes. Journal of artificial intelligence research, 2, pp.263-286.](https://arxiv.org/pdf/cs/9501101.pdf)

    """

    def __init__(self, classifier: base.BinaryClassifier, code_size: int, seed: int = None):
        self.classifier = classifier
        self.code_size = code_size
        self.seed = seed
        self._rng = random.Random(seed)

        self.classifiers = {i: copy.deepcopy(classifier) for i in range(code_size)}

        # We don't how many classes there are, therefore we can't generate the code book
        # from the start. Therefore, we define a random queue of integers. When a new class
        # appears, we get the next integer and convert it to a code.
        integers = list(range(2 ** code_size))
        self._rng.shuffle(integers)
        self._ints = iter(integers)

        def next_code():
            i = next(self._ints)
            b = bin(i)[2:]  # convert to a string of 0s and 1s
            b = b.zfill(self.code_size)  # ensure the code is of length code_size
            return tuple(int(c) for c in b)

        self.code_book = collections.defaultdict(next_code)

    @property
    def _wrapped_model(self):
        return self.classifier

    def fit_one(self, x, y):

        code = self.code_book[y]

        for i, c in enumerate(code):
            self.classifiers[i].fit_one(x, c)

        return self

    def predict_one(self, x):

        if not self.code_book:  # it's empty
            return None

        output = [None for _ in range(self.code_size)]

        for i, clf in self.classifiers.items():
            output[i] = clf.predict_proba_one(x).get(True, 0.)

        return min(self.code_book, key=lambda c: l1_dist(self.code_book[c], output))



import abc

from . import estimator


class AnomalyDetector(estimator.Estimator):

    @property
    def _is_supervised(self):
        return False

    @abc.abstractmethod
    def fit_one(self, x: dict) -> 'AnomalyDetector':
        """Update the model.

        Parameters:
            x: A dictionary of features.

        Returns:
            self

        """

    @abc.abstractmethod
    def score_one(self, x: dict) -> float:
        """Return an outlier score.

        A high score is indicative of an anomaly. A low score indicates a normal observation.

        Parameters:
            x: A dictionary of features.

        Returns:
            The anomaly score.

        """



import glob
import inspect
import os
import pickle
import platform

from creme import utils


class Cache:
    """Utility for caching iterables.

    This can be used to save a stream of data to the disk in order to iterate over it faster the
    following time. This can save time depending on the nature of stream. The more processing
    happens in a stream, the more time will be saved. Even in the case where no processing is done
    apart from reading the data, the cache will save some time because it is using the pickle
    binary protocol. It can thus improve the speed in common cases such as reading from a CSV file.

    Parameters:
        directory (str): The path where to store the pickled data streams. If not provided, then
            it will be automatically inferred whenever possible, if not an exception will be
            raised.

    Attributes:
        keys (set): The set of keys that are being cached.

    Example:

        >>> import time
        >>> from creme import datasets
        >>> from creme import stream

        >>> X_y = datasets.Phishing()
        >>> cache = stream.Cache()

        The cache can be used by wrapping it around an iterable. Because this is the first time
        are iterating over the data, nothing is cached.

        >>> tic = time.time()
        >>> for x, y in cache(X_y, key='phishing'):
        ...     pass
        >>> toc = time.time()
        >>> print(toc - tic)  # doctest: +SKIP
        0.012813

        If we do the same thing again, we can see the loop is now faster.

        >>> tic = time.time()
        >>> for x, y in cache(X_y, key='phishing'):
        ...     pass
        >>> toc = time.time()
        >>> print(toc - tic)  # doctest: +SKIP
        0.001927

        We can see an overview of the cache. The first line indicates the location of the
        cache.

        >>> cache  # doctest: +SKIP
        /tmp
        phishing - 125.2KiB

        Finally, we can clear the stream from the cache.

        >>> cache.clear('phishing')
        >>> cache
        /tmp

        There is also a ``clear_all`` method to remove all the items in the cache.

        >>> cache.clear_all()

    """

    def __init__(self, directory=None):

        # Guess the directory from the system
        system = platform.system()
        if directory is None:
            directory = {'Linux': '/tmp', 'Darwin': '/tmp'}.get(system)

        if directory is None:
            raise ValueError('There is no default directory defined for {systems} systems, '
                             'please provide one')

        self.directory = directory
        self.keys = set()

        # Check if there is anything already in the cache
        for f in glob.glob(f'{self.directory}/*.creme_cache.pkl'):
            key = os.path.basename(f).split('.')[0]
            self.keys.add(key)

    def _get_path(self, key):
        return os.path.join(self.directory, f'{key}.creme_cache.pkl')

    def __call__(self, stream, key=None):

        # Try to guess a key from the stream object
        if key is None:
            if inspect.isfunction(stream):
                key = stream.__name__

        if key is None:
            raise ValueError('No default key could be guessed for the given stream, '
                             'please provide one')

        path = self._get_path(key)

        if os.path.exists(path):
            yield from self[key]
            return

        with open(path, 'wb') as f:
            pickler = pickle.Pickler(f)
            for el in stream:
                pickler.dump(el)
                yield el
            self.keys.add(key)

    def __getitem__(self, key):
        """Iterates over the stream associated with the given key."""
        with open(self._get_path(key), 'rb') as f:
            unpickler = pickle.Unpickler(f)
            while f.peek(1):
                yield unpickler.load()

    def clear(self, key: str):
        """Deletes the cached stream associated with the given key.

        Parameters:
            key

        """
        os.remove(self._get_path(key))
        self.keys.remove(key)

    def clear_all(self):
        """Deletes all the cached streams."""
        for key in list(self.keys):
            os.remove(self._get_path(key))
            self.keys.remove(key)

    def __repr__(self):
        return '\n'.join([self.directory] + [
            f'{key} - {utils.pretty.humanize_bytes(os.path.getsize(self._get_path(key)))}'
            for key in self.keys
        ])



from scipy import integrate

from . import base
from . import confusion


__all__ = ['ROCAUC']


class ROCAUC(base.BinaryMetric):
    """Receiving Operating Characteristic Area Under the Curve.

    This metric is an approximation of the true ROC AUC. Computing the true ROC AUC would
    require storing all the predictions and true outputs, which isn't an option. The approximation
    error is not significant as long as the predicted probabilities are well calibrated. In any
    case, this metric can still be used to reliably compare models, even though it is an
    approximation.

    Parameters:
        n_thresholds: The number of thresholds to use to discretize the ROC curve. The higher this
            is, the closer the output will be to the true ROC AUC value, at the cost of more time
            and memory.

    Example:

        >>> from creme import metrics

        >>> y_true = [ 0,  0,   1,  1]
        >>> y_pred = [.1, .4, .35, .8]

        >>> metric = metrics.ROCAUC()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     metric = metric.update(yt, yp)

        >>> metric
        ROCAUC: 0.875

        The true ROC AUC is in fact 0.75. We can improve the accuracy by increasing the amount
        of thresholds. This comes at the cost more computation time and more memory usage.

        >>> metric = metrics.ROCAUC(n_thresholds=20)

        >>> for yt, yp in zip(y_true, y_pred):
        ...     metric = metric.update(yt, yp)

        >>> metric
        ROCAUC: 0.75

    """

    def __init__(self, n_thresholds=10):
        self.n_thresholds = n_thresholds
        self.thresholds = [i / (n_thresholds - 1) for i in range(n_thresholds)]
        """The equidistant thresholds."""
        self.thresholds[0] -= 1e-7
        self.thresholds[-1] += 1e-7
        self.cms = [confusion.ConfusionMatrix() for _ in range(n_thresholds)]
        """Contains a `metrics.ConfusionMatrix` for each threshold."""

    def update(self, y_true, y_pred, sample_weight=1.):
        p_true = y_pred.get(True, 0.) if isinstance(y_pred, dict) else y_pred
        for t, cm in zip(self.thresholds, self.cms):
            cm.update(y_true=bool(y_true), y_pred=p_true > t, sample_weight=sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        p_true = y_pred.get(True, 0.) if isinstance(y_pred, dict) else y_pred
        for t, cm in zip(self.thresholds, self.cms):
            cm.revert(y_true=bool(y_true), y_pred=p_true > t, sample_weight=sample_weight)
        return self

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return False

    def get(self):

        tprs = [0] * self.n_thresholds
        fprs = [0] * self.n_thresholds

        def safe_div(a, b):
            try:
                return a / b
            except ZeroDivisionError:
                return 0.

        for i, cm in enumerate(self.cms):
            tp = cm.counts.get(True, {}).get(True, 0)
            tn = cm.counts.get(False, {}).get(False, 0)
            fp = cm.counts.get(False, {}).get(True, 0)
            fn = cm.counts.get(True, {}).get(False, 0)

            tprs[i] = safe_div(a=tp, b=tp + fn)
            fprs[i] = safe_div(a=fp, b=fp + tn)

        return -integrate.trapz(x=fprs, y=tprs)



from . import base
from . import count


class CentralMoments(base.Univariate):
    """Computes central moments using Welford's algorithm.

    Attributes:
        count (stats.Count)
        delta (float): Mean of differences.
        sum_delta (float): Mean of sum of differences.
        M1 (float): sums of powers of differences from the mean order 1.
        M2 (float): sums of powers of differences from the mean order 2.
        M3 (float): sums of powers of differences from the mean order 3.
        M4 (float): sums of powers of differences from the mean order 4.

    References:
        1. [Wikipedia article on algorithms for calculating variance](https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Covariance)

    """

    def __init__(self):
        self.count = count.Count()

        self.delta = 0
        self.sum_delta = 0

        self.M1 = 0
        self.M2 = 0
        self.M3 = 0
        self.M4 = 0

    def _update_delta(self, x):
        self.delta = (x - self.sum_delta) / self.count.get()
        return self

    def _update_sum_delta(self):
        self.sum_delta += self.delta
        return self

    def _update_m1(self, x):
        self.M1 = (x - self.sum_delta) * self.delta * (self.count.get() - 1)
        return self

    def _update_m2(self):
        self.M2 += self.M1
        return self

    def _update_m3(self):
        self.M3 += (self.M1 * self.delta * (self.count.get() - 2) - 3 *
                    self.delta * self.M2)
        return self

    def _update_m4(self):
        delta_square = self.delta ** 2
        self.M4 += (self.M1 * delta_square *
                    (self.count.get() ** 2 - 3 * self.count.get() + 3) +
                    6 * delta_square * self.M2 - 4 * self.delta * self.M3)
        return self



import os

from creme import stream

from . import base


class ChickWeights(base.FileDataset):
    """Chick weights along time.

    The stream contains 578 items and 3 features. The goal is to predict the weight of each chick
    along time, according to the diet the chick is on. The data is ordered by time and then by
    chick.

    References:
        1. [Chick weight dataset overview](http://rstudio-pubs-static.s3.amazonaws.com/107631_131ad1c022df4f90aa2d214a5c5609b2.html)

    """

    def __init__(self):
        super().__init__(filename='chick-weights.csv', n_samples=578, n_features=3, task=base.REG)

    def __iter__(self):
        return stream.iter_csv(
            self.path,
            target='weight',
            converters={'time': int, 'weight': int, 'chick': int, 'diet': int}
        )



from creme import stream

from . import base


class Higgs(base.RemoteDataset):
    """Higgs dataset.

    The data has been produced using Monte Carlo simulations. The first 21 features (columns 2-22)
    are kinematic properties measured by the particle detectors in the accelerator. The last seven
    features are functions of the first 21 features; these are high-level features derived by
    physicists to help discriminate between the two classes.

    References:
        1. [UCI page](https://archive.ics.uci.edu/ml/datasets/HIGGS)

    """

    def __init__(self):
        super().__init__(
            n_samples=11_000_000,
            n_features=28,
            task=base.BINARY_CLF,
            url='https://archive.ics.uci.edu/ml/machine-learning-databases/00280/HIGGS.csv.gz',
            size=2740166144,
            unpack=False
        )

    def _iter(self):

        features = [
            'lepton pT', 'lepton eta', 'lepton phi',
            'missing energy magnitude', 'missing energy phi',
            'jet 1 pt', 'jet 1 eta', 'jet 1 phi', 'jet 1 b-tag',
            'jet 2 pt', 'jet 2 eta', 'jet 2 phi', 'jet 2 b-tag',
            'jet 3 pt', 'jet 3 eta', 'jet 3 phi', 'jet 3 b-tag',
            'jet 4 pt', 'jet 4 eta', 'jet 4 phi', 'jet 4 b-tag',
            'm_jj', 'm_jjj', 'm_lv', 'm_jlv', 'm_bb', 'm_wbb', 'm_wwbb'
        ]

        return stream.iter_csv(
            self.path,
            fieldnames=['is_signal', *features],
            target='is_signal',
            converters={'is_signal': lambda x: x.startswith('1'), **{f: float for f in features}}
        )



import numbers
import functools
import collections

import numpy as np
import pandas as pd

from creme import base
from creme import stats
from creme import utils


__all__ = [
    'Binarizer',
    'MaxAbsScaler',
    'MinMaxScaler',
    'Normalizer',
    'RobustScaler',
    'StandardScaler'
]


def safe_div(a, b):
    """Returns a if b is nil, else divides a by b.

    When scaling, sometimes a denominator might be nil. For instance, during standard scaling
    the denominator can be nil if a feature has no variance.

    """
    return a / b if b else 0.


class Binarizer(base.Transformer):
    """Binarizes the data to 0 or 1 according to a threshold.

    Parameters:
        threshold: Values above this are replaced by 1 and the others by 0.
        dtype: The desired data type to apply.

    Example:

        >>> import creme
        >>> import numpy as np

        >>> rng = np.random.RandomState(42)
        >>> X = [{'x1': v, 'x2': int(v)} for v in rng.uniform(low=-4, high=4, size=6)]

        >>> binarizer = creme.preprocessing.Binarizer()
        >>> for x in X:
        ...     print(binarizer.fit_one(x).transform_one(x))
        {'x1': False, 'x2': False}
        {'x1': True, 'x2': True}
        {'x1': True, 'x2': True}
        {'x1': True, 'x2': False}
        {'x1': False, 'x2': False}
        {'x1': False, 'x2': False}

    """

    def __init__(self, threshold=0., dtype=bool):
        self.threshold = threshold
        self.dtype = dtype

    def transform_one(self, x):
        x_tf = x.copy()

        for i, xi in x_tf.items():
            if isinstance(xi, numbers.Number):
                x_tf[i] = self.dtype(xi > self.threshold)

        return x_tf


class StandardScaler(base.Transformer):
    """Scales the data so that it has zero mean and unit variance.

    Under the hood, a running mean and a running variance are maintained. The scaling is slightly
    different than when scaling the data in batch because the exact means and variances are not
    known in advance. However, this doesn't have a detrimental impact on performance in the long
    run.

    This transformer supports mini-batches as well as single instances. In the mini-batch case, the
    number of columns and the ordering of the columns are allowed to change between subsequent
    calls. In other words, this transformer will keep working even if you add and/or remove
    features every time you call `fit_many` and `transform_many`.

    Example:

        >>> from pprint import pprint
        >>> import random
        >>> from creme import preprocessing

        >>> random.seed(42)
        >>> X = [{'x': random.uniform(8, 12), 'y': random.uniform(8, 12)} for _ in range(6)]
        >>> pprint(X)
        [{'x': 10.557, 'y': 8.100},
         {'x': 9.100, 'y': 8.892},
         {'x': 10.945, 'y': 10.706},
         {'x': 11.568, 'y': 8.347},
         {'x': 9.687, 'y': 8.119},
         {'x': 8.874, 'y': 10.021}]

        >>> scaler = preprocessing.StandardScaler()

        >>> for x in X:
        ...     print(scaler.fit_one(x).transform_one(x))
        {'x': 0.0, 'y': 0.0}
        {'x': -0.999, 'y': 0.999}
        {'x': 0.937, 'y': 1.350}
        {'x': 1.129, 'y': -0.651}
        {'x': -0.776, 'y': -0.729}
        {'x': -1.274, 'y': 0.992}

        This transformer also supports mini-batch updates. You can call `fit_many` and provide a
        `pandas.DataFrame`:

        >>> import pandas as pd
        >>> X = pd.DataFrame.from_dict(X)

        >>> scaler = preprocessing.StandardScaler()
        >>> scaler = scaler.fit_many(X[:3])
        >>> scaler = scaler.fit_many(X[3:])

        You can then call `transform_many` to scale a mini-batch of features:

        >>> scaler.transform_many(X)
           x         y
        0  0.444600 -0.933384
        1 -1.044259 -0.138809
        2  0.841106  1.679208
        3  1.477301 -0.685117
        4 -0.444084 -0.914195
        5 -1.274664  0.992296

    References:
        1. [Welford's Method (and Friends)](https://www.embeddedrelated.com/showarticle/785.php)
        2. [Batch updates for simple statistics](https://notmatthancock.github.io/2017/03/23/simple-batch-stat-updates.html)

    """

    def __init__(self):
        self.counts = collections.Counter()
        self.means = collections.defaultdict(float)
        self.vars = collections.defaultdict(float)

    def fit_one(self, x):

        for i, xi in x.items():
            self.counts[i] += 1
            old_mean = self.means[i]
            self.means[i] += (xi - old_mean) / self.counts[i]
            self.vars[i] += ((xi - old_mean) * (xi - self.means[i]) - self.vars[i]) / self.counts[i]

        return self

    def transform_one(self, x):
        return {
            i: safe_div(xi - self.means[i], self.vars[i] ** .5)
            for i, xi in x.items()
        }

    def fit_many(self, X: pd.DataFrame):
        """Update with a mini-batch of features.

        Note that the update formulas for mean and variance are slightly different than in the
        single instance case, but they produce exactly the same result.

        Parameters:
            X: A dataframe where each column is a feature.

        """

        # Operating on X.values, which is view to the underlying numpy array, is slightly faster
        # than operating on X
        columns = X.columns
        X = X.values

        # In the rest of this method, old_* refers to the existing statistics, whilst new_* refers
        # to the statistics of the current mini-batch.

        new_means = np.nanmean(X, axis=0)
        # We could call np.var, but we already have the mean so we can be smart
        new_vars = np.einsum('ij,ij->j', X, X) / len(X) - new_means ** 2
        new_counts = np.sum(~np.isnan(X), axis=0)

        for col, new_mean, new_var, new_count in zip(columns, new_means, new_vars, new_counts):

            old_mean = self.means[col]
            old_var = self.vars[col]
            old_count = self.counts[col]

            a = old_count / (old_count + new_count)
            b = new_count / (old_count + new_count)

            self.means[col] = a * old_mean + b * new_mean
            self.vars[col] = (
                a * old_var +
                b * new_var +
                a * b * (old_mean - new_mean) ** 2
            )
            self.counts[col] += new_count

        return self

    def transform_many(self, X):
        """Scale a mini-batch of features.

        Parameters:
            X: A dataframe where each column is a feature. An exception will be raised if any of
                the features has not been seen during a previous call to `fit_many`.

        """

        means = np.array([self.means[c] for c in X.columns])
        stds = np.array([self.vars[c] ** .5 for c in X.columns])

        return pd.DataFrame(
            np.divide((X.values - means), stds, where=stds > 0),
            index=X.index,
            columns=X.columns,
            copy=False
        )


class MinMaxScaler(base.Transformer):
    """Scales the data to a fixed range from 0 to 1.

    Under the hood a running min and a running peak to peak (max - min) are maintained.

    Attributes:
        min (dict): Mapping between features and instances of `stats.Min`.
        max (dict): Mapping between features and instances of `stats.Max`.

    Example:

        >>> from pprint import pprint
        >>> import random
        >>> from creme import preprocessing

        >>> random.seed(42)
        >>> X = [{'x': random.uniform(8, 12)} for _ in range(5)]
        >>> pprint(X)
        [{'x': 10.557707},
         {'x': 8.100043},
         {'x': 9.100117},
         {'x': 8.892842},
         {'x': 10.945884}]

        >>> scaler = preprocessing.MinMaxScaler()

        >>> for x in X:
        ...     print(scaler.fit_one(x).transform_one(x))
        {'x': 0.0}
        {'x': 0.0}
        {'x': 0.406920}
        {'x': 0.322582}
        {'x': 1.0}

    """

    def __init__(self):
        self.min = collections.defaultdict(stats.Min)
        self.max = collections.defaultdict(stats.Max)

    def fit_one(self, x):

        for i, xi in x.items():
            self.min[i].update(xi)
            self.max[i].update(xi)

        return self

    def transform_one(self, x):
        return {
            i: safe_div(xi - self.min[i].get(),
                        self.max[i].get() - self.min[i].get())
            for i, xi in x.items()
        }


class MaxAbsScaler(base.Transformer):
    """Scales the data to a [-1, 1] range based on absolute maximum.

    Under the hood a running absolute max is maintained. This scaler is meant for
    data that is already centered at zero or sparse data. It does not shift/center
    the data, and thus does not destroy any sparsity.

    Attributes:
        abs_max (dict): Mapping between features and instances of `stats.AbsMax`.

    Example:

        >>> from pprint import pprint
        >>> import random
        >>> from creme import preprocessing

        >>> random.seed(42)
        >>> X = [{'x': random.uniform(8, 12)} for _ in range(5)]
        >>> pprint(X)
        [{'x': 10.557707},
         {'x': 8.100043},
         {'x': 9.100117},
         {'x': 8.892842},
         {'x': 10.945884}]

        >>> scaler = preprocessing.MaxAbsScaler()

        >>> for x in X:
        ...     print(scaler.fit_one(x).transform_one(x))
        {'x': 1.0}
        {'x': 0.767216}
        {'x': 0.861940}
        {'x': 0.842308}
        {'x': 1.0}

    """

    def __init__(self):
        self.abs_max = collections.defaultdict(stats.AbsMax)

    def fit_one(self, x):

        for i, xi in x.items():
            self.abs_max[i].update(xi)

        return self

    def transform_one(self, x):
        return {
            i: safe_div(xi, self.abs_max[i].get())
            for i, xi in x.items()
        }


class RobustScaler(base.Transformer):
    """Scale features using statistics that are robust to outliers.

    This Scaler removes the median and scales the data according to the
    interquantile range.

    Parameters:
        with_centering: Whether to centre the data before scaling.
        with_scaling: Whether to scale data to IQR.
        q_inf: Desired inferior quantile, must be between 0 and 1.
        q_sup: Desired superior quantile, must be between 0 and 1.

    Attributes:
        median (dict): Mapping between features and instances of `stats.Quantile(0.5)`.
        iqr (dict): Mapping between features and instances of `stats.IQR`.

    Example:

        >>> from pprint import pprint
        >>> import random
        >>> from creme import preprocessing

        >>> random.seed(42)
        >>> X = [{'x': random.uniform(8, 12)} for _ in range(5)]
        >>> pprint(X)
        [{'x': 10.557707},
         {'x': 8.100043},
         {'x': 9.100117},
         {'x': 8.892842},
         {'x': 10.945884}]

        >>> scaler = preprocessing.RobustScaler()

        >>> for x in X:
        ...     print(scaler.fit_one(x).transform_one(x))
        {'x': 0.0}
        {'x': -1.0}
        {'x': 0.0}
        {'x': -0.124499}
        {'x': 1.108659}

    """

    def __init__(self, with_centering=True, with_scaling=True, q_inf=.25, q_sup=.75):
        self.with_centering = with_centering
        self.with_scaling = with_scaling
        self.q_inf = q_inf
        self.q_sup = q_sup
        self.median = collections.defaultdict(functools.partial(stats.Quantile, 0.5))
        self.iqr = collections.defaultdict(functools.partial(stats.IQR, self.q_inf, self.q_sup))

    def fit_one(self, x):

        for i, xi in x.items():
            if self.with_centering:
                self.median[i].update(xi)
            if self.with_scaling:
                self.iqr[i].update(xi)

        return self

    def transform_one(self, x):
        x_tf = {}

        for i, xi in x.items():
            x_tf[i] = xi
            if self.with_centering:
                x_tf[i] -= self.median[i].get()
            if self.with_scaling:
                x_tf[i] = safe_div(x_tf[i], self.iqr[i].get())

        return x_tf


class Normalizer(base.Transformer):
    """Scales a set of features so that it has unit norm.

    This is particularly useful when used after a `feature_extraction.TFIDF`.

    Parameters:
        order: Order of the norm (e.g. 2 corresponds to the $L^2$ norm).

    Example:

        >>> from creme import preprocessing
        >>> from creme import stream

        >>> scaler = preprocessing.Normalizer(order=2)

        >>> X = [[4, 1, 2, 2],
        ...      [1, 3, 9, 3],
        ...      [5, 7, 5, 1]]

        >>> for x, _ in stream.iter_array(X):
        ...     print(scaler.transform_one(x))
        {0: 0.8, 1: 0.2, 2: 0.4, 3: 0.4}
        {0: 0.1, 1: 0.3, 2: 0.9, 3: 0.3}
        {0: 0.5, 1: 0.7, 2: 0.5, 3: 0.1}

    """

    def __init__(self, order=2):
        self.order = order

    def transform_one(self, x):
        norm = utils.math.norm(x, order=self.order)
        return {i: xi / norm for i, xi in x.items()}



import collections
import functools
import io
import itertools
import types
import typing
import warnings

import pandas as pd

from .. import base

from . import func
from . import union


__all__ = ['Pipeline']


class Pipeline(base.Estimator):
    """A pipeline of estimators.

    Pipelines allow you to chain different steps into a sequence. Typically, when doing supervised
    learning, a pipeline contains one ore more transformation steps, whilst it's is a regressor or
    a classifier. It is highly recommended to use pipelines with `creme`. Indeed, in an online
    learning setting, it is very practical to have a model defined as a single object. Take a look
    at the [user guide](/user-guide/the-art-of-using-pipelines) for further information and
    practical examples.

    One special thing to take notice to is the way transformers are handled. In a typical scenario,
    it is usual to predict something for a sample and wait for the ground truth to arrive. In such
    a case, the features are seen before the ground truth arrives. Therefore, the unsupervised
    parts of the pipeline are updated when `predict_one` and `predict_proba_one` are called.
    Usually the unsupervised parts of the pipeline are all the steps that precede the final step,
    which is a supervised model. However, some transformers are supervised and are therefore
    obtained during calls to `fit_one`.

    Parameters:
        steps: Ideally, a list of (name, estimator) tuples. A name is automatically inferred
            if none is provided.

    Example:

        >>> from creme import compose
        >>> from creme import linear_model
        >>> from creme import preprocessing

        The recommended way to declare a pipeline is to use the `|` operator. The latter allows you
        to chain estimators in a very terse manner:

        >>> from creme import linear_model
        >>> from creme import preprocessing

        >>> scaler = preprocessing.StandardScaler()
        >>> log_reg = linear_model.LinearRegression()
        >>> model = scaler | log_reg

        This results in a pipeline that stores each step inside a dictionary.

        >>> model
        Pipeline (
          StandardScaler (),
          LinearRegression (
            optimizer=SGD (
              lr=Constant (
                learning_rate=0.01
              )
            )
            loss=Squared ()
            l2=0.
            intercept=0.
            intercept_lr=Constant (
            learning_rate=0.01
            )
            clip_gradient=1e+12
            initializer=Zeros ()
          )
        )

        You can access parts of a pipeline in the same manner as a dictionary:

        >>> model['LinearRegression']
        LinearRegression (
          optimizer=SGD (
            lr=Constant (
              learning_rate=0.01
            )
          )
          loss=Squared ()
          l2=0.
          intercept=0.
          intercept_lr=Constant (
            learning_rate=0.01
          )
          clip_gradient=1e+12
          initializer=Zeros ()
        )

        Note that you can also declare a pipeline by using the `compose.Pipeline` constructor
        method, which is slighly more verbose:

        >>> from creme import compose

        >>> model = compose.Pipeline(scaler, log_reg)

        By using a `compose.TransformerUnion`, you can define complex pipelines that apply
        different steps to different parts of the data. For instance, we can extract word counts
        from text data, and extract polynomial features from numeric data.

        >>> from creme import feature_extraction

        >>> tfidf = feature_extraction.TFIDF('text')
        >>> counts = feature_extraction.BagOfWords('text')
        >>> text_part = compose.Select('text') | (tfidf + counts)

        >>> num_part = compose.Select('a', 'b') | preprocessing.PolynomialExtender()

        >>> model = text_part + num_part
        >>> model |= preprocessing.StandardScaler()
        >>> model |= linear_model.LinearRegression()

        You can obtain a visual representation of the pipeline by calling it's `draw` method.

        >>> dot = model.draw()

        ![pipeline_example](/img/pipeline_docstring.svg)

        The following shows an example of using `debug_one` to visualize how the information
        flows and changes throughout the pipeline.

        >>> from creme import compose
        >>> from creme import feature_extraction
        >>> from creme import naive_bayes

        >>> X_y = [
        ...     ('A positive comment', True),
        ...     ('A negative comment', False),
        ...     ('A happy comment', True),
        ...     ('A lovely comment', True),
        ...     ('A harsh comment', False)
        ... ]

        >>> tfidf = feature_extraction.TFIDF() | compose.Renamer(prefix='tfidf_')
        >>> counts = feature_extraction.BagOfWords() | compose.Renamer(prefix='count_')
        >>> mnb = naive_bayes.MultinomialNB()
        >>> model = (tfidf + counts) | mnb

        >>> for x, y in X_y:
        ...     model = model.fit_one(x, y)

        >>> x = X_y[0][0]
        >>> report = model.debug_one(X_y[0][0])
        >>> print(report)
        0. Input
        --------
        A positive comment
        <BLANKLINE>
        1. Transformer union
        --------------------
            1.0 TFIDF | Renamer
            -------------------
            tfidf_comment: 0.47606 (float)
            tfidf_positive: 0.87942 (float)
        <BLANKLINE>
            1.1 BagOfWords | Renamer
            ------------------------
            count_comment: 1 (int)
            count_positive: 1 (int)
        <BLANKLINE>
        count_comment: 1 (int)
        count_positive: 1 (int)
        tfidf_comment: 0.50854 (float)
        tfidf_positive: 0.86104 (float)
        <BLANKLINE>
        2. MultinomialNB
        ----------------
        False: 0.19313
        True: 0.80687

    """

    def __init__(self, *steps):
        self.steps = collections.OrderedDict()
        for step in steps:
            self |= step

    def __getitem__(self, key):
        """Just for convenience."""
        return self.steps[key]

    def __len__(self):
        """Just for convenience."""
        return len(self.steps)

    def __or__(self, other):
        """Inserts a step at the end of the pipeline."""
        self._add_step(other, at_start=False)
        return self

    def __ror__(self, other):
        """Inserts a step at the start of the pipeline."""
        self._add_step(other, at_start=True)
        return self

    def __add__(self, other):
        """Merges with another Pipeline or TransformerUnion into a TransformerUnion."""
        if isinstance(other, union.TransformerUnion):
            return other.__add__(self)
        return union.TransformerUnion(self, other)

    def __str__(self):
        return ' | '.join(map(str, self.steps.values()))

    def __repr__(self):
        return (
            'Pipeline (\n\t' +
            '\t'.join(',\n'.join(map(repr, self.steps.values())).splitlines(True)) +
            '\n)'
        ).expandtabs(2)

    def _get_params(self):
        return dict(self.steps.items())

    def _set_params(self, new_params=None):
        if new_params is None:
            new_params = {}
        return Pipeline(*[
            (name, new_params[name])
            if isinstance(new_params.get(name), base.Estimator) else
            (name, step._set_params(new_params.get(name, {})))
            for name, step in self.steps.items()
        ])

    @property
    def _is_supervised(self):
        return any(step._is_supervised for step in self.steps.values())

    def _add_step(self, estimator, at_start: bool):
        """Adds a step to either end of the pipeline.

        This method takes care of sanitizing the input. For instance, if a function is passed,
        then it will be wrapped with a `FuncTransformer`.

        """

        name = None
        if isinstance(estimator, tuple):
            name, estimator = estimator

        # If the step is a function then wrap it in a FuncTransformer
        if isinstance(estimator, (types.FunctionType, types.LambdaType)):
            estimator = func.FuncTransformer(estimator)

        def infer_name(estimator):
            if isinstance(estimator, func.FuncTransformer):
                return infer_name(estimator.func)
            elif isinstance(estimator, (types.FunctionType, types.LambdaType)):
                return estimator.__name__
            elif hasattr(estimator, '__class__'):
                return estimator.__class__.__name__
            return str(estimator)

        # Infer a name if none is given
        if name is None:
            name = infer_name(estimator)

        if name in self.steps:
            counter = 1
            while f'{name}{counter}' in self.steps:
                counter += 1
            name = f'{name}{counter}'

        # Instantiate the estimator if it hasn't been done
        if isinstance(estimator, type):
            estimator = estimator()

        # Store the step
        self.steps[name] = estimator

        # Move the step to the start of the pipeline if so instructed
        if at_start:
            self.steps.move_to_end(name, last=False)

    # Single instance methods

    def fit_one(self, x: dict, y=None, **params):
        """Fit to a single instance.

        Parameters:
            X: A dictionary of features.
            y: A target value.

        """

        steps = iter(self.steps.values())

        # Loop over the first n - 1 steps, which should all be transformers
        for t in itertools.islice(steps, len(self.steps) - 1):
            x_pre = x
            x = t.transform_one(x=x)

            # The supervised transformers have to be updated.
            # Note that this is done after transforming in order to avoid target leakage.
            if isinstance(t, union.TransformerUnion):
                for sub_t in t.transformers.values():
                    if sub_t._is_supervised:
                        sub_t.fit_one(x=x_pre, y=y)

            elif t._is_supervised:
                t.fit_one(x=x_pre, y=y)

        # At this point steps contains a single step, which is therefore the final step of the
        # pipeline
        final = next(steps)
        if final._is_supervised:
            final.fit_one(x=x, y=y, **params)
        else:
            final.fit_one(x=x, **params)

        return self

    def _transform_one(self, x: dict):
        """This methods takes care of applying the first n - 1 steps of the pipeline, which are
        supposedly transformers. It also returns the final step so that other functions can do
        something with it.

        """

        steps = iter(self.steps.values())

        for t in itertools.islice(steps, len(self.steps) - 1):

            # The unsupervised transformers are updated during transform. We do this because
            # typically transform_one is called before fit_one, and therefore we might as well use
            # the available information as soon as possible. Note that way of proceeding is very
            # specific to online machine learning.
            if isinstance(t, union.TransformerUnion):
                for sub_t in t.transformers.values():
                    if not sub_t._is_supervised:
                        sub_t.fit_one(x=x)

            elif not t._is_supervised:
                t.fit_one(x=x)

            x = t.transform_one(x=x)

        return x, next(steps)

    def transform_one(self, x: dict):
        """Apply each transformer in the pipeline to some features.

        The final step in the pipeline will be applied if it is a transformer. If not, then it will
        be ignored and the output from the penultimate step will be returned. Note that the steps
        that precede the final step are assumed to all be transformers.

        """
        x, final_step = self._transform_one(x=x)
        if isinstance(final_step, base.Transformer):
            return final_step.transform_one(x=x)
        return x

    def predict_one(self, x: dict):
        x, final_step = self._transform_one(x=x)
        return final_step.predict_one(x=x)

    def predict_proba_one(self, x: dict):
        x, final_step = self._transform_one(x=x)
        return final_step.predict_proba_one(x=x)

    def score_one(self, x: dict):
        x, final_step = self._transform_one(x=x)
        return final_step.score_one(x=x)

    def forecast(self, horizon: int, xs: typing.List[dict] = None):
        """Return a forecast.

        Only works if each estimator has a `transform_one` method and the final estimator has a
        `forecast` method. This is the case of time series models from the `time_series` module.

        Parameters:
            horizon: The forecast horizon.
            xs: A list of features for each step in the horizon.

        """
        if xs is not None:
            xs = [self._transform_one(x)[0] for x in xs]
        final_step = list(self.steps.values())[-1]
        return final_step.forecast(horizon=horizon, xs=xs)

    def debug_one(self, x: dict, show_types=True, n_decimals=5) -> str:
        """Displays the state of a set of features as it goes through the pipeline.

        Parameters:
            x A set of features.
            show_types: Whether or not to display the type of feature along with it's value.
            n_decimals: Number of decimals to display for each floating point value.

        """

        tab = ' ' * 4

        # We'll redirect all the print statement to a buffer, we'll return the content of the
        # buffer at the end
        buffer = io.StringIO()
        _print = functools.partial(print, file=buffer)

        def format_value(x):
            if isinstance(x, float):
                return '{:,.{prec}f}'.format(x, prec=n_decimals)
            return x

        def print_dict(x, show_types, indent=False, space_after=True):

            # Some transformers accept strings as input instead of dicts
            if isinstance(x, str):
                _print(x)
            else:
                for k, v in sorted(x.items()):
                    type_str = f' ({type(v).__name__})' if show_types else ''
                    _print((tab if indent else '') + f'{k}: {format_value(v)}' + type_str)
            if space_after:
                _print()

        def print_title(title, indent=False):
            _print((tab if indent else '') + title)
            _print((tab if indent else '') + '-' * len(title))

        # Print the initial state of the features
        print_title('0. Input')
        print_dict(x, show_types=show_types)

        # Print the state of x at each step
        steps = iter(self.steps.values())
        for i, t in enumerate(itertools.islice(steps, len(self.steps) - 1)):

            if isinstance(t, union.TransformerUnion):
                print_title(f'{i+1}. Transformer union')
                for j, (name, sub_t) in enumerate(t.transformers.items()):
                    if isinstance(sub_t, Pipeline):
                        name = str(sub_t)
                    print_title(f'{i+1}.{j} {name}', indent=True)
                    print_dict(sub_t.transform_one(x), show_types=show_types, indent=True)
                x = t.transform_one(x)
                print_dict(x, show_types=show_types)

            else:
                print_title(f'{i+1}. {t}')
                x = t.transform_one(x)
                print_dict(x, show_types=show_types)

        # Print the predicted output from the final estimator
        final = next(steps)
        if not isinstance(final, base.Transformer):
            print_title(f'{len(self)}. {final}')

            # If the last estimator has a debug_one method then call it
            if hasattr(final, 'debug_one'):
                _print(final.debug_one(x))

            # Display the prediction
            _print()
            if isinstance(final, base.Classifier):
                print_dict(final.predict_proba_one(x), show_types=False, space_after=False)
            else:
                _print(f'Prediction: {format_value(final.predict_one(x))}')

        return buffer.getvalue().rstrip()

    # Mini-batch methods

    def fit_many(self, X: pd.DataFrame, y: pd.Series, **params):
        """Fit to a mini-batch.

        Parameters:
            X: A dataframe of features. Columns can be added and/or removed between successive
                calls.
            y: A series of target values.

        """

        steps = iter(self.steps.values())

        # Loop over the first n - 1 steps, which should all be transformers
        for t in itertools.islice(steps, len(self.steps) - 1):
            X_pre = X
            X = t.transform_many(X=X)

            # The supervised transformers have to be updated.
            # Note that this is done after transforming in order to avoid target leakage.
            if isinstance(t, union.TransformerUnion):
                for sub_t in t.transformers.values():
                    if isinstance(sub_t, base.SupervisedTransformer):
                        sub_t.fit_many(X=X)

            elif isinstance(t, base.SupervisedTransformer):
                t.fit_many(X=X)

        # At this point steps contains a single step, which is therefore the final step of the
        # pipeline
        final = next(steps)

        if final._is_supervised:
            final.fit_many(X=X, y=y, **params)
        else:
            final.fit_many(X=X, y=y, **params)

        return self

    def _transform_many(self, X: pd.DataFrame):
        """This methods takes care of applying the first n - 1 steps of the pipeline, which are
        supposedly transformers. It also returns the final step so that other functions can do
        something with it.

        """

        steps = iter(self.steps.values())

        for t in itertools.islice(steps, len(self.steps) - 1):

            # The unsupervised transformers are updated during transform. We do this because
            # typically transform_one is called before fit_one, and therefore we might as well use
            # the available information as soon as possible. Note that way of proceeding is very
            # specific to online machine learning.
            if isinstance(t, union.TransformerUnion):
                for sub_t in t.transformers.values():
                    if not isinstance(t, base.SupervisedTransformer):
                        sub_t.fit_many(X=X)

            elif not isinstance(t, base.SupervisedTransformer):
                t.fit_many(X=X)

            X = t.transform_many(X=X)

        return X, next(steps)

    def transform_many(self, X: pd.DataFrame):
        """Apply each transformer in the pipeline to some features.

        The final step in the pipeline will be applied if it is a transformer. If not, then it will
        be ignored and the output from the penultimate step will be returned. Note that the steps
        that precede the final step are assumed to all be transformers.

        """
        X, final_step = self._transform_many(X=X)
        if isinstance(final_step, base.Transformer):
            return final_step.transform_many(X=X)
        return X

    def predict_many(self, X: pd.DataFrame):
        X, final_step = self._transform_many(X=X)
        return final_step.predict_many(X=X)

    def predict_proba_many(self, X: pd.DataFrame):
        X, final_step = self._transform_many(X=X)
        return final_step.predict_proba_many(X=X)

    def draw(self):
        """Draws the pipeline using the `graphviz` library."""

        def networkify(step):

            # Unions are converted to an undirected network
            if isinstance(step, union.TransformerUnion):
                return Network(
                    nodes=map(networkify, step.transformers.values()),
                    edges=[],
                    directed=False
                )

            # Pipelines are converted to a directed network
            if isinstance(step, Pipeline):
                return Network(
                    nodes=[],
                    edges=zip(
                        map(networkify, list(step.steps.values())[:-1]),
                        map(networkify, list(step.steps.values())[1:])
                    ),
                    directed=True
                )

            # Wrapper models are handled recursively
            if isinstance(step, base.Wrapper):
                return Network(
                    nodes=[networkify(step._wrapped_model)],
                    edges=[],
                    directed=True,
                    name=type(step).__name__,
                    labelloc=step._labelloc
                )

            # Other steps are treated as strings
            return str(step)

        # Draw input
        net = Network(nodes=['x'], edges=[], directed=True)
        previous = 'x'

        # Draw each step
        for step in self.steps.values():
            current = networkify(step)
            net.link(previous, current)
            previous = current

        # Draw output
        net.link(previous, 'y')

        return net.draw()


class Network(collections.UserList):
    """An abstraction to help with drawing pipelines."""

    def __init__(self, nodes, edges, directed, name=None, labelloc=None):
        super().__init__()
        for node in nodes:
            self.append(node)
        self.edges = set()
        for link in edges:
            self.link(*link)
        self.directed = directed
        self.name = name
        self.labelloc = labelloc

    def append(self, a):
        if a not in self:
            super().append(a)

    def link(self, a, b):
        self.append(a)
        self.append(b)
        self.edges.add((self.index(a), self.index(b)))

    def draw(self):

        import graphviz

        G = graphviz.Digraph()

        drawn_subclusters = set()

        def draw_node(a):
            if isinstance(a, Network):
                for part in a:
                    draw_node(part)
            else:
                G.node(a)

        for a in self:
            draw_node(a)

        def draw_link(a, b):

            if isinstance(a, Network):

                # Connect the last part of a with b
                if a.directed:
                    draw_link(a[-1], b)
                # Connect each part of a with b
                else:
                    for part in a:
                        draw_link(part, b)

            elif isinstance(b, Network):

                # Connect the first part of b with a
                if b.directed:

                    if str(b) not in drawn_subclusters:

                        sub = b.draw()

                        # If the graph has a name, then we treat is as a cluster
                        if b.name is not None:
                            sub.attr(label=b.name, labelloc=b.labelloc)
                            sub.name = f'cluster_{b.name}'

                        G.subgraph(sub)

                        drawn_subclusters.add(str(b))

                    draw_link(a, b[0])

                # Connect each part of b with a
                else:
                    for part in b:
                        draw_link(a, part)

            else:
                G.edge(a, b)

        for a, b in self.edges:
            draw_link(self[a], self[b])

        return G



from creme import stream

from . import base


class Bananas(base.FileDataset):
    """Bananas dataset.

    An artificial dataset where instances belongs to several clusters with a banana shape.
    There are two attributes that correspond to the x and y axis, respectively.

    References:
        1. [OpenML page](https://www.openml.org/d/1460)

    """

    def __init__(self):
        super().__init__(filename='banana.zip', n_samples=5300, n_features=2, task=base.BINARY_CLF)

    def __iter__(self):
        return stream.iter_libsvm(self.path, target_type=lambda x: x == '1')



import math

from . import rmse


__all__ = ['RMSLE']


class RMSLE(rmse.RMSE):
    """Root mean squared logarithmic error.

    Example:

        >>> from creme import metrics

        >>> y_true = [3, -0.5, 2, 7]
        >>> y_pred = [2.5, 0.0, 2, 8]

        >>> metric = metrics.RMSLE()
        >>> for yt, yp in zip(y_true, y_pred):
        ...     metric = metric.update(yt, yp)

        >>> metric
        RMSLE: 0.357826

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        return super().update(math.log(y_true + 1), math.log(y_pred + 1), sample_weight)



import numbers
import typing

FeatureName = typing.Hashable
RegTarget = numbers.Number
ClfTarget = typing.Union[bool, str, int]
Target = typing.Union[ClfTarget, RegTarget]
Stream = typing.Iterator[typing.Tuple[dict, typing.Any]]



import collections

from . import base


__all__ = ['Adam']


class Adam(base.Optimizer):
    """Adam optimizer.

    Parameters:
        lr
        beta_1
        beta_2
        eps

    Attributes:
        m (collections.defaultdict)
        v (collections.defaultdict)

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.Adam()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.86496

    References:
        1. [Kingma, D.P. and Ba, J., 2014. Adam: A method for stochastic optimization. arXiv preprint arXiv:1412.6980.](https://arxiv.org/pdf/1412.6980.pdf)

    """

    def __init__(self, lr=0.1, beta_1=0.9, beta_2=0.999, eps=1e-8):
        super().__init__(lr)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.eps = eps
        self.m = collections.defaultdict(float)
        self.v = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        # Correct bias for `v`
        lr = self.learning_rate * (1 - self.beta_2 ** (self.n_iterations + 1)) ** .5
        # Correct bias for `m`
        lr /= (1 - self.beta_1 ** (self.n_iterations + 1))

        for i, gi in g.items():
            self.m[i] = self.beta_1 * self.m[i] + (1 - self.beta_1) * gi
            self.v[i] = self.beta_2 * self.v[i] + (1 - self.beta_2) * gi ** 2
            w[i] -= lr * self.m[i] / (self.v[i] ** .5 + self.eps)

        return w



import abc
import itertools
import pathlib
import os
import shutil
import tarfile
from urllib import request
import zipfile

from creme import utils


REG = 'Regression'
BINARY_CLF = 'Binary classification'
MULTI_CLF = 'Multi-class classification'
MO_BINARY_CLF = 'Multi-output binary classification'


def get_data_home():
    """Return the location where remote datasets are to be stored.

    """

    data_home = os.environ.get('CREME_DATA', os.path.join('~', 'creme_data'))
    data_home = os.path.expanduser(data_home)
    if not os.path.exists(data_home):
        os.makedirs(data_home)
    return data_home


class Dataset(abc.ABC):
    """Base class for all datasets.

    All datasets inherit from this class, be they stored in a file or generated on the fly.

    """

    def __init__(self, task, n_features, n_samples=None, n_classes=None, sparse=False):
        self.task = task
        self.n_features = n_features
        self.n_samples = n_samples
        self.n_classes = n_classes
        self.sparse = sparse

    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError

    def take(self, k: int):
        """Iterate over the k samples."""
        return itertools.islice(self, k)

    @property
    def _repr_title(self):
        return f'{self.__class__.__name__} dataset'

    @property
    def _repr_content(self):
        """The items that are displayed in the __repr__ method.

        This property can be overriden in order to modify the output of the __repr__ method.

        """

        content = {}
        content['Task'] = self.task
        if self.n_samples:
            content['Number of samples'] = f'{self.n_samples:,}'
        if self.n_features:
            content['Number of features'] = f'{self.n_features:,}'
        if self.n_classes:
            content['Number of classes'] = f'{self.n_classes:,}'
        content['Sparse'] = str(self.sparse)

        return content

    def __repr__(self):

        l_len = max(map(len, self._repr_content.keys()))
        r_len = max(map(len, self._repr_content.values()))

        return (
            f'{self._repr_title}\n\n' +
            '\n'.join(
                k.rjust(l_len) + '  ' + v.ljust(r_len)
                for k, v in self._repr_content.items()
            )
        )


class SyntheticDataset(Dataset):

    @property
    def _repr_title(self):
        return f'{self.__class__.__name__} synthetic dataset'



class FileDataset(Dataset):
    """Base class for datasets that are stored in a local file.

    Small datasets that are part of the creme package inherit from this class.

    """

    def __init__(self, filename, **desc):
        super().__init__(**desc)
        self.filename = filename

    @property
    def path(self):
        return pathlib.Path(__file__).parent.joinpath(self.filename)

    @property
    def _repr_content(self):
        content = super()._repr_content
        content['Path'] = str(self.path)
        return content


class RemoteDataset(FileDataset):
    """Base class for datasets that are stored in a remote file.

    Medium and large datasets that are not part of the creme package inherit from this class.

    The filename doesn't have to be provided if unpack is False. Indeed in the latter case the
    filename will be inferred from the URL.

    """

    def __init__(self, url, size, filename=None, unpack=True, **desc):

        if filename is None:
            filename = os.path.basename(url)

        super().__init__(filename=filename, **desc)
        self.url = url
        self.size = size
        self.unpack = unpack

    @property
    def path(self):
        return pathlib.Path(get_data_home(), self.__class__.__name__, self.filename)

    def download(self, verbose=True):

        # Determine where to download the archive
        directory = self.path.parent
        directory.mkdir(parents=True, exist_ok=True)
        archive_path = directory.joinpath(os.path.basename(self.url))

        with request.urlopen(self.url) as r:

            # Notify the user
            if verbose:
                meta = r.info()
                try:
                    n_bytes = int(meta['Content-Length'])
                    msg = f'Downloading {self.url} ({utils.pretty.humanize_bytes(n_bytes)})'
                except KeyError:
                    msg = f'Downloading {self.url}'
                print(msg)

            # Now dump the contents of the requests
            with open(archive_path, 'wb') as f:
                shutil.copyfileobj(r, f)

        if not self.unpack:
            return

        if verbose:
            print(f'Uncompressing into {directory}')

        if archive_path.suffix.endswith('zip'):
            with zipfile.ZipFile(archive_path, 'r') as zf:
                zf.extractall(directory)

        elif archive_path.suffix.endswith(('gz', 'tar')):
            mode = 'r:' if archive_path.suffix.endswith('tar') else 'r:gz'
            tar = tarfile.open(archive_path, mode)
            tar.extractall(directory)
            tar.close()

        else:
            raise RuntimeError(f'Unhandled extension type: {archive_path.suffix}')

        # Delete the archive file now that it has been uncompressed
        archive_path.unlink()

    @abc.abstractmethod
    def _iter(self):
        pass

    @property
    def is_downloaded(self):
        """Indicate whether or the data has been correctly downloaded."""
        if self.path.exists():

            if self.path.is_file():
                return self.path.stat().st_size == self.size
            return sum(f.stat().st_size for f in self.path.glob('**/*') if f.is_file())

        return False

    def __iter__(self):
        if not self.is_downloaded:
            self.download(verbose=True)
        if not self.is_downloaded:
            raise RuntimeError('Something went wrong during the download')
        yield from self._iter()

    @property
    def _repr_content(self):
        content = super()._repr_content
        content['URL'] = self.url
        content['Size'] = utils.pretty.humanize_bytes(self.size)
        content['Downloaded'] = str(self.is_downloaded)
        return content



from . import base
from . import ewmean


class EWVar(base.Univariate):
    """Exponentially weighted variance.

    To calculate the variance we use the fact that Var(X) = Mean(x^2) - Mean(x)^2 and internally
    we use the exponentially weighted mean of x/x^2 to calculate this.

    Parameters:
        alpha: The closer `alpha` is to 1 the more the statistic will adapt to recent values.

    Attributes:
        variance (float): The running exponentially weighted variance.

    Example:

        >>> from creme import stats

        >>> X = [1, 3, 5, 4, 6, 8, 7, 9, 11]
        >>> ewv = stats.EWVar(alpha=0.5)
        >>> for x in X:
        ...     print(ewv.update(x).get())
        0
        1.0
        2.75
        1.4375
        1.984375
        3.43359375
        1.7958984375
        2.198974609375
        3.56536865234375

    References:
        1. [Finch, T., 2009. Incremental calculation of weighted mean and variance. University of Cambridge, 4(11-5), pp.41-42.](http://people.ds.cam.ac.uk/fanf2/hermes/doc/antiforgery/stats.pdf)
        2. [Exponential Moving Average on Streaming Data](https://dev.to/nestedsoftware/exponential-moving-average-on-streaming-data-4hhl)

    """

    def __init__(self, alpha=0.5):
        self.alpha = alpha
        self.mean = ewmean.EWMean(alpha=alpha)
        self.sq_mean = ewmean.EWMean(alpha=alpha)

    @property
    def name(self):
        return f'ewv_{self.alpha}'

    def update(self, x):
        self.mean.update(x)
        self.sq_mean.update(x ** 2)
        return self

    def get(self):
        return self.sq_mean.get() - self.mean.get() ** 2



cdef class Statistic:
    cpdef double get(self)

cdef class Univariate(Statistic):
    cpdef Univariate update(self, double x)
    cpdef Univariate revert(self, double x)

cdef class Bivariate(Statistic):
    cpdef Bivariate update(self, double x, double y)



import math

from creme import proba


def entropy(dist: proba.Multinomial):
    """Returns the entropy of a multinomial distribution.

    Parameters:
        dist

    Example:

        >>> from creme import proba

        >>> events = [
        ...     'sunny', 'sunny',
        ...     'rainy', 'rainy', 'rainy',
        ...     'snowy', 'snomy', 'snowy', 'snomy', 'snowy'
        ... ]

        >>> dist = proba.Multinomial(events)

        >>> entropy(dist)
        1.97095

    References:
        1. [A Simple Explanation of Information Gain and Entropy](https://victorzhou.com/blog/information-gain/)
        2. [Calculating entropy <https://www.johndcook.com/blog/2013/08/17/calculating-entropy/)

    """
    return -sum(dist.pmf(c) * math.log2(dist.pmf(c)) for c in dist if dist.pmf(c) > 0)


def gini_impurity(dist: proba.Multinomial):
    """Returns the Gini impurity of a multinomial distribution.

    Parameters:
        dist

    Example:

        >>> from creme import proba

        >>> events = [
        ...     'sunny', 'sunny',
        ...     'rainy', 'rainy', 'rainy',
        ...     'snowy', 'snomy', 'snowy', 'snomy', 'snowy'
        ... ]

        >>> dist = proba.Multinomial(events)

        >>> gini_impurity(dist)
        0.74

    References:
        1. [A Simple Explanation of Gini Impurity](https://victorzhou.com/blog/gini-impurity/)

    """
    return sum(dist.pmf(c) * (1 - dist.pmf(c)) for c in dist)



"""Factorization machines."""
from .ffm import FFMClassifier
from .ffm import FFMRegressor
from .fm import FMClassifier
from .fm import FMRegressor
from .fwfm import FwFMClassifier
from .fwfm import FwFMRegressor
from .hofm import HOFMClassifier
from .hofm import HOFMRegressor


__all__ = [
    'FFMClassifier',
    'FFMRegressor',
    'FMClassifier',
    'FMRegressor',
    'FwFMClassifier',
    'FwFMRegressor',
    'HOFMClassifier',
    'HOFMRegressor',
]



from creme import stats

from . import base


__all__ = ['Jaccard']


class Jaccard(base.MeanMetric, base.MultiOutputClassificationMetric):
    """Jaccard index for binary multi-outputs.

    Example:

        >>> from creme import metrics

        >>> y_true = [
        ...     {0: False, 1: True, 2: True},
        ...     {0: True, 1: True, 2: False},
        ... ]

        >>> y_pred = [
        ...     {0: True, 1: True, 2: True},
        ...     {0: True, 1: False, 2: False},
        ... ]

        >>> jac = metrics.Jaccard()
        >>> for yt, yp in zip(y_true, y_pred):
        ...     jac = jac.update(yt, yp)

        >>> jac
        Jaccard: 0.583333

    References:
        1. [Wikipedia section on similarity of asymmetric binary attributes](https://www.wikiwand.com/en/Jaccard_index#/Similarity_of_asymmetric_binary_attributes)

    """

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def _eval(self, y_true, y_pred):

        one_and_ones = 0
        zero_or_ones = 0

        for i in y_true:
            if y_true[i] and y_pred[i]:
                one_and_ones += 1
            elif y_true[i] or y_pred[i]:
                zero_or_ones += 1

        return one_and_ones / (zero_or_ones + one_and_ones)



import typing

import pandas as pd

from creme import base
from creme import stream


def iter_pandas(X: pd.DataFrame, y: typing.Union[pd.Series, pd.DataFrame] = None,
                **kwargs) -> base.typing.Stream:
    """Iterates over the rows of a `pandas.DataFrame`.

    Parameters:
        X: A dataframe of features.
        y: A series or a dataframe with one column per target.

    Example:

        >>> import pandas as pd
        >>> from creme import stream

        >>> X = pd.DataFrame({
        ...     'x1': [1, 2, 3, 4],
        ...     'x2': ['blue', 'yellow', 'yellow', 'blue'],
        ...     'y': [True, False, False, True]
        ... })
        >>> y = X.pop('y')

        >>> for xi, yi in stream.iter_pandas(X, y):
        ...     print(xi, yi)
        {'x1': 1, 'x2': 'blue'} True
        {'x1': 2, 'x2': 'yellow'} False
        {'x1': 3, 'x2': 'yellow'} False
        {'x1': 4, 'x2': 'blue'} True

    """

    kwargs['feature_names'] = X.columns
    if isinstance(y, pd.DataFrame):
        kwargs['target_names'] = y.columns

    yield from stream.iter_array(X=X.to_numpy(), y=y if y is None else y.to_numpy(), **kwargs)



import ast

from creme import stream

from . import base


class Restaurants(base.RemoteDataset):
    """Data from the Kaggle Recruit Restaurants challenge.

    The goal is to predict the number of visitors in each of 829 Japanese restaurants over a priod
    of roughly 16 weeks. The data is ordered by date and then by restaurant ID.

    References:
        1. [Recruit Restaurant Visitor Forecasting](https://www.kaggle.com/c/recruit-restaurant-visitor-forecasting)

    """

    def __init__(self):
        super().__init__(
            n_samples=252_108,
            n_features=7,
            task=base.REG,
            url='https://maxhalford.github.io/files/datasets/kaggle_recruit_restaurants.zip',
            size=28881242,
            filename='kaggle_recruit_restaurants.csv'
        )

    def _iter(self):
        return stream.iter_csv(
            self.path,
            target='visitors',
            converters={
                'latitude': float,
                'longitude': float,
                'visitors': int,
                'is_holiday': ast.literal_eval
            },
            parse_dates={'date': '%Y-%m-%d'}
        )



"""Multi-class classification."""
from .ooc import OutputCodeClassifier
from .ovo import OneVsOneClassifier
from .ovr import OneVsRestClassifier


__all__ = [
    'OutputCodeClassifier',
    'OneVsOneClassifier',
    'OneVsRestClassifier'
]



import collections
import copy
import functools
import typing

from creme import base


__all__ = ['Grouper']


class Grouper(base.Transformer):
    """Applies a transformer within different groups.

    This transformer allows you to split your data into groups and apply a transformer within each
    group. This happens in a streaming manner, which means that the groups are discovered online.
    A separate copy of the provided transformer is made whenever a new group appears. The groups
    are defined according to one or more keys.

    Parameters:
        transformer
        by: The field on which to group the data. This can either by a single value, or a list of
            values.

    """

    def __init__(self, transformer: base.Transformer,
                 by: typing.Union[base.typing.FeatureName, typing.List[base.typing.FeatureName]]):

        self.transformer = transformer
        self.by = by if isinstance(by, list) else [by]
        self.transformers = collections.defaultdict(functools.partial(copy.deepcopy, transformer))

    def _get_key(self, x):
        return '_'.join(str(x[k]) for k in self.by)

    def fit_one(self, x):
        key = self._get_key(x)
        self.transformers[key].fit_one(x)
        return self

    def transform_one(self, x):
        key = self._get_key(x)
        return self.transformers[key].transform_one(x)



import random

from creme import stats

from . import base


__all__ = ['RandomNormal']


class RandomNormal(base.Recommender):
    """Predicts random values sampled from a normal distribution.

    The parameters of the normal distribution are fitted with running statistics. This is
    equivalent to using `surprise.prediction_algorithms.random_pred.NormalPredictor`. The model
    expect dict inputs containing both a `user` and an `item` entries.

    Parameters:
        seed: Randomization seed used for reproducibility.

    Attributes:
        mean (stats.Mean)
        variance (stats.Var)

    Example:

        >>> from creme import reco

        >>> X_y = (
        ...     ({'user': 'Alice', 'item': 'Superman'}, 8),
        ...     ({'user': 'Alice', 'item': 'Terminator'}, 9),
        ...     ({'user': 'Alice', 'item': 'Star Wars'}, 8),
        ...     ({'user': 'Alice', 'item': 'Notting Hill'}, 2),
        ...     ({'user': 'Alice', 'item': 'Harry Potter'}, 5),
        ...     ({'user': 'Bob', 'item': 'Superman'}, 8),
        ...     ({'user': 'Bob', 'item': 'Terminator'}, 9),
        ...     ({'user': 'Bob', 'item': 'Star Wars'}, 8),
        ...     ({'user': 'Bob', 'item': 'Notting Hill'}, 2)
        ... )

        >>> model = reco.RandomNormal(seed=42)

        >>> for x, y in X_y:
        ...     _ = model.fit_one(x, y)

        >>> model.predict_one({'user': 'Bob', 'item': 'Harry Potter'})
        6.883895

    Note:
        This model expects a dict input with a `user` and an `item` entries without any type
        constraint on their values (i.e. can be strings or numbers). Other entries are ignored.

    """

    def __init__(self, seed=None):
        super().__init__()
        self.variance = stats.Var()
        self.mean = stats.Mean()
        self.seed = seed
        self._rng = random.Random(seed)

    def _fit_one(self, user, item, y):
        y_pred = self._predict_one(user, item)
        self.mean.update(y)
        self.variance.update(y)
        return y_pred

    def _predict_one(self, user, item):
        μ = self.mean.get() or 0
        σ = (self.variance.get() or 1) ** 0.5
        return self._rng.gauss(μ, σ)



import collections
import heapq
import itertools
import math


__all__ = ['Histogram']


class Bin:
    """A Bin is an element of a Histogram."""

    __slots__ = ['left', 'right', 'count']

    def __init__(self, left, right, count):
        self.left = left
        self.right = right
        self.count = count

    def __iadd__(self, other):
        """Merge with another bin."""
        if other.left < self.left:
            self.left = other.left
        if other.right > self.right:
            self.right = other.right
        self.count += other.count
        return self

    def __lt__(self, other):
        return self.right < other.left

    def __eq__(self, other):
        return self.left == other.left and self.right == other.right

    def __repr__(self):
        return f'[{self.left:.5f}, {self.right:.5f}]: {self.count}'


def coverage_ratio(x: Bin, y: Bin) -> float:
    """Returns the amount of y covered by x.

    Examples:

        >>> coverage_ratio(Bin(1, 2, 0), Bin(1, 2, 0))
        1.0

        >>> coverage_ratio(Bin(1, 3, 0), Bin(2, 4, 0))
        0.5

        >>> coverage_ratio(Bin(1, 3, 0), Bin(3, 5, 0))
        0.0

        >>> coverage_ratio(Bin(1, 3, 0), Bin(0, 4, 0))
        0.5

        >>> coverage_ratio(Bin(0, 4, 0), Bin(1, 3, 0))
        1.0

        >>> coverage_ratio(Bin(1, 3, 0), Bin(0, 1, 0))
        0.0

        >>> coverage_ratio(Bin(1, 1, 0), Bin(1, 1, 0))
        1.0

    """
    if y.left == y.right:
        return float(x.left <= y.left <= x.right)
    return max(0, min(x.right, y.right) - max(x.left, y.left)) / (y.right - y.left)


class Histogram(collections.UserList):
    """Streaming histogram.

    Parameters:
        max_bins: Maximal number of bins.

    Attributes:
        n (int): Total number of seen values.

    Example:

        >>> from creme import utils
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np

        >>> np.random.seed(42)

        >>> values = np.hstack((
        ...     np.random.normal(-3, 1, 1000),
        ...     np.random.normal(3, 1, 1000),
        ... ))

        >>> hist = utils.Histogram(max_bins=60)

        >>> for x in values:
        ...     hist = hist.update(x)

        >>> ax = plt.bar(
        ...     x=[(b.left + b.right) / 2 for b in hist],
        ...     height=[b.count for b in hist],
        ...     width=[(b.right - b.left) / 2 for b in hist]
        ... )

    .. image:: ../../docs/img/histogram_docstring.svg
        :align: center

    References:
        1. [Ben-Haim, Y. and Tom-Tov, E., 2010. A streaming parallel decision tree algorithm. Journal of Machine Learning Research, 11(Feb), pp.849-872.](http://jmlr.org/papers/volume11/ben-haim10a/ben-haim10a.pdf)
        2. [Go implementation](https://github.com/VividCortex/gohistogram)

    """

    def __init__(self, max_bins=256):
        super().__init__()
        self.max_bins = max_bins
        self.n = 0

    def update(self, x):

        self.n += 1
        b = Bin(x, x, 1)

        # Insert the bin if the histogram is empty
        if not self:
            self.append(b)
            return self

        # Use bisection to find where to insert
        # We don't use the bisect module in order to save some CPU cycles
        lo = 0
        hi = len(self)
        i = (lo + hi) // 2
        while lo < hi:
            if self[i] < b:
                lo = i + 1
            else:
                hi = i
            i = (lo + hi) // 2

        if i == len(self):
            # x is past the right-most bin
            self.append(b)
        else:
            # Increment the bin counter if x is part of the ith bin
            if x >= self[i].left:
                self[i].count += 1
            # Insert the bin if it is between bin i-1 and bin i
            else:
                self.insert(i, b)

        # Bins have to be merged if there are more than max_bins
        if len(self) > self.max_bins:
            self._shrink(1)

        return self

    def _shrink(self, k):
        """Shrinks the histogram by merging the two closest bins."""

        if k == 1:

            # Find the closest pair of bins
            min_diff = math.inf
            min_idx = None
            for idx, (b1, b2) in enumerate(zip(self.data[:-1], self.data[1:])):
                diff = b2.right - b1.right
                if diff < min_diff:
                    min_diff = diff
                    min_idx = idx

            # Merge the bins
            self[min_idx] += self.pop(min_idx + 1)
            return

        indexes = range(len(self) - 1)

        def bin_distance(i):
            return self[i + 1].right - self[i].right

        for i in sorted(heapq.nsmallest(n=k, iterable=indexes, key=bin_distance), reverse=True):
            self[i] += self.pop(i + 1)  # Calls Bin.__iadd__

    def iter_cdf(self, X, verbose=False):
        """Yields CDF values for a sorted iterable of values.

        This is faster than calling `cdf` with many values.

        Example:

            >>> from creme import utils

            >>> hist = utils.Histogram()
            >>> for x in range(4):
            ...     hist = hist.update(x)

            >>> print(hist)
            [0.00000, 0.00000]: 1
            [1.00000, 1.00000]: 1
            [2.00000, 2.00000]: 1
            [3.00000, 3.00000]: 1

            >>> X = [-1, 0, .5, 1, 2.5, 3.5]
            >>> for x, cdf in zip(X, hist.iter_cdf(X)):
            ...     print(x, cdf)
            -1 0.0
            0 0.25
            0.5 0.25
            1 0.5
            2.5 0.75
            3.5 1.0

        """

        bins = iter(self)
        b = next(bins)
        INF = Bin(math.inf, math.inf, 0)

        cdf = 0

        for x in X:

            while x >= b.right:
                cdf += b.count
                b = next(bins, INF)

            if x > b.left:
                yield (cdf + b.count * (x - b.left) / (b.right - b.left)) / self.n
                continue

            yield cdf / self.n

    def cdf(self, x):
        """Cumulative distribution function.

        Example:

            >>> from creme import utils

            >>> hist = utils.Histogram()
            >>> for x in range(4):
            ...     hist = hist.update(x)

            >>> print(hist)
            [0.00000, 0.00000]: 1
            [1.00000, 1.00000]: 1
            [2.00000, 2.00000]: 1
            [3.00000, 3.00000]: 1

            >>> hist.cdf(-1)
            0.0

            >>> hist.cdf(0)
            0.25

            >>> hist.cdf(.5)
            0.25

            >>> hist.cdf(1)
            0.5

            >>> hist.cdf(2.5)
            0.75

            >>> hist.cdf(3.5)
            1.0

        """
        return next(self.iter_cdf([x]))

    def __add__(self, other):
        """

        Example:

            >>> h1 = Histogram()
            >>> for b in [Bin(0, 2, 4), Bin(4, 5, 9)]:
            ...     h1.append(b)

            >>> h2 = Histogram()
            >>> for b in [Bin(1, 3, 8), Bin(3, 4, 5)]:
            ...     h2.append(b)

            >>> h1 + h2
            [0.00000, 1.00000]: 2.0
            [1.00000, 2.00000]: 6.0
            [2.00000, 3.00000]: 4.0
            [3.00000, 4.00000]: 5.0
            [4.00000, 5.00000]: 9.0

        """

        xs = iter(self)
        ys = iter(other)
        rights = heapq.merge(
            itertools.chain.from_iterable((b.left, b.right) for b in self),
            itertools.chain.from_iterable((b.left, b.right) for b in other)
        )

        b = Bin(next(rights), next(rights), 0)

        x = next(xs)
        y = next(ys)

        hist = Histogram(max_bins=max(self.max_bins, other.max_bins))

        while True:

            b.count += coverage_ratio(b, x) * x.count
            b.count += coverage_ratio(b, y) * y.count

            if b.count:
                hist.append(b)

            try:
                b = Bin(b.right, next(rights), 0)
            except StopIteration:
                break

            if b.left >= x.right:
                x = next(xs, x)
            if b.left >= y.right:
                y = next(ys, y)

        hist._shrink(k=len(hist) - hist.max_bins)

        return hist

    def __repr__(self):
        return '\n'.join(str(b) for b in self)



from creme import stream

from . import base


class Phishing(base.FileDataset):
    """Phishing websites.

    This dataset contains features from web pages that are classified as phishing or not.

    References:
        1. [UCI page](http://archive.ics.uci.edu/ml/datasets/Website+Phishing)

    """

    def __init__(self):
        super().__init__(
            n_samples=1250,
            n_features=9,
            task=base.BINARY_CLF,
            filename='phishing.csv.gz'
        )

    def __iter__(self):
        return stream.iter_csv(
            self.path,
            target='is_phishing',
            converters={
                'empty_server_form_handler': float,
                'popup_window': float,
                'https': float,
                'request_from_other_domain': float,
                'anchor_from_other_domain': float,
                'is_popular': float,
                'long_url': float,
                'age_of_domain': int,
                'ip_in_url': int,
                'is_phishing': lambda x: x == '1'
            }
        )



import collections
import copy

from creme import base
from creme import utils


__all__ = ['OneVsRestClassifier']


class OneVsRestClassifier(base.Wrapper, base.MultiClassifier):
    """One-vs-the-rest (OvR) multiclass strategy.

    This strategy consists in fitting one binary classifier per class. Because we are in a
    streaming context, the number of classes isn't known from the start, hence new classifiers are
    instantiated on the fly. Likewise, the predicted probabilities will only include the classes
    seen up to a given point in time.

    Parameters:
        classifier: A binary classifier, although a multi-class classifier will work too.

    Attributes:
        classifiers (dict): A mapping between classes and classifiers.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import multiclass
        >>> from creme import preprocessing

        >>> X_y = datasets.ImageSegments()

        >>> scaler = preprocessing.StandardScaler()
        >>> ovr = multiclass.OneVsRestClassifier(linear_model.LogisticRegression())
        >>> model = scaler | ovr

        >>> metric = metrics.MacroF1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MacroF1: 0.774573

    """

    def __init__(self, classifier: base.BinaryClassifier):
        self.classifier = classifier
        self.classifiers = {}

    @property
    def _wrapped_model(self):
        return self.classifier

    def fit_one(self, x, y):

        # Instantiate a new binary classifier if the class is new
        if y not in self.classifiers:
            self.classifiers[y] = copy.deepcopy(self.classifier)

        # Train each label's associated classifier
        for label, model in self.classifiers.items():
            model.fit_one(x, y == label)

        return self

    def predict_proba_one(self, x):
        y_pred = {
            label: model.predict_proba_one(x)[True]
            for label, model in self.classifiers.items()
        }
        return utils.math.softmax(y_pred)
